from __future__ import annotations

from collections import defaultdict
import functools
import re
import sys
from typing import TYPE_CHECKING

from .. import exc
from ..exc import Base
from ..exc import BaseError
from ..exc import BaseWarning
from ..exc import ErrorCompilingKernel

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Sequence

    from .source_location import SourceLocation
    from .type_propagation import TypeNotAllowedOnDevice
    from helion.runtime.settings import Settings

    ErrorOrWarning = BaseError | BaseWarning


# pyre-ignore[9]
order_by_location: Callable[[Sequence[ErrorOrWarning]], list[ErrorOrWarning]] = (
    functools.partial(sorted, key=lambda e: e.location)
)


class ErrorReporting:
    def __init__(self, settings: Settings) -> None:
        """
        Initialize the ErrorReporting object with the provided settings.

        :param settings: The Settings object containing configuration for error reporting.
        """
        self.errors: list[BaseError] = []
        self.warnings: list[BaseWarning] = []
        self.ignores: tuple[type[BaseWarning], ...] = tuple(settings.ignore_warnings)
        self.type_errors: dict[SourceLocation, list[exc.TypePropagationError]] = (
            defaultdict(list)
        )
        self.printed_warning = 0

    def add(self, e: Base | type[Base]) -> None:
        """
        Add an error or warning to the respective list.

        :param e: An instance or type of BaseError or BaseWarning.
        :raises TypeError: If the provided object is not an error or warning.
        """
        if callable(e):
            e = e()
        if isinstance(e, BaseError):
            self.errors.append(e)
        elif isinstance(e, BaseWarning):
            if not isinstance(e, self.ignores):
                self.warnings.append(e)
        else:
            raise TypeError(f"expected error or warning, got {type(e)}")

    def add_type_error(self, type_info: TypeNotAllowedOnDevice) -> None:
        """
        Add a type error to the list of type errors and to the errors list if it's the first occurrence.

        :param type_info: The TypeNotAllowedOnDevice object containing information about the type error.
        """
        locations = type_info.locations
        similar_errors = self.type_errors[locations[0]]
        similar_errors.append(e := exc.TypePropagationError(type_info, similar_errors))
        if len(similar_errors) == 1:
            self.add(e)

    def ignore(self, e: type[BaseWarning]) -> None:
        """
        Add a warning type to the ignore list.

        :param e: The type of BaseWarning to ignore.
        """
        self.ignores = (*self.ignores, e)

    def raise_if_errors(self) -> None:
        """
        Raise an exception if there are any errors, after reporting pending warnings.

        :raises ErrorCompilingKernel: If there are multiple errors.
        :raises BaseError: If there is a single error.
        """
        sys.stderr.write(self.report(warnings_offset=self.printed_warning))
        self.printed_warning = len(self.warnings)
        if self.errors:
            if len(self.errors) > 1:
                raise ErrorCompilingKernel(len(self.errors), len(self.warnings))
            raise self.errors[0]

    def report(self, *, strip_paths: bool = False, warnings_offset: int = 0) -> str:
        """
        Generate a report of the errors and warnings.

        :param strip_paths: Whether to strip file paths from the report.
        :param warnings_offset: The offset to start reporting warnings from.
        :return: A string representation of the report.
        """
        report = "\n\n".join(
            {
                e.report(): None
                for e in order_by_location(
                    [*self.errors, *self.warnings[warnings_offset:]]
                )
            }
        )
        if strip_paths:
            report = re.sub(r'"[/\\][^"]*[/\\]', '".../', report)
        return report



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import NamedTuple

import torch

from .. import exc
from .._compiler.ast_extension import expr_from_string
from . import _decorators

if TYPE_CHECKING:
    import ast

    from .._compiler.inductor_lowering import CodegenState
    from .._compiler.type_propagation import TypeInfo
    from .._compiler.variable_origin import Origin


class ConstExpr(NamedTuple):
    """
    Typically used as a type annotation for kernels:

        @helion.kernel()
        def fn(v: hl.constexpr, ...):
            ...

    Causes the generated code to specialize on the value of `v`, where a different
    kernel, hardcoding the value of v, will be generated every time `v` changes.
    """

    value: object

    def __index__(self) -> int:
        if isinstance(self.value, int):
            return self.value
        raise TypeError(f"ConstExpr cannot be indexed: {self.value}")

    def __bool__(self) -> bool:
        return bool(self.value)


@_decorators.api(is_device_only=False)
def specialize(value: int | torch.SymInt) -> int:
    """
    Turn a dynamic shape into a compile-time constant.  Example:

           hl.specialize(tensor.size(1))

    :param value: The symbolic value to specialize on.
    :return: The specialized value.
    """
    raise exc.NotInsideKernel


@_decorators.type_propagation(specialize)
def _(value: TypeInfo, *, origin: Origin) -> TypeInfo:
    from .._compiler.compile_environment import CompileEnvironment
    from .._compiler.type_propagation import TypeInfo

    if origin.is_device():
        raise exc.SpecializeOnDevice
    proxy = value.proxy()
    if isinstance(proxy, torch.SymInt):
        CompileEnvironment.current().specialized_vars.update(
            proxy._sympy_().free_symbols
        )
        return TypeInfo.from_example(proxy.__int__(), origin=origin)
    if isinstance(proxy, int):
        return TypeInfo.from_example(proxy, origin=origin)  # already specialized
    raise exc.SpecializeArgType(value)


# pyre-fixme[56]
@_decorators.codegen(specialize)
def _(state: CodegenState) -> ast.AST:
    value = state.proxy_arg(0)
    if isinstance(value, torch.SymInt):
        value = value.__int__()
    assert isinstance(value, int)
    return expr_from_string(repr(value))



from __future__ import annotations

import ast
import collections
import typing
from typing import TYPE_CHECKING
from typing import TypeVar

if TYPE_CHECKING:
    _A = TypeVar("_A", bound=ast.AST)


# TODO(oulgen): This visitor is extremely primitive, does not consider alpha renaming or scopes
class _ReadWriteVisitor(ast.NodeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.rw = ReadWrites(collections.Counter(), collections.Counter())

    def _update(self, name: str, ctx: ast.expr_context) -> None:
        if isinstance(ctx, ast.Load):
            self.rw.reads[name] += 1
        elif isinstance(ctx, ast.Store):
            self.rw.writes[name] += 1

    def visit_Name(self, node: ast.Name) -> None:
        self._update(node.id, node.ctx)

    def visit_Subscript(self, node: ast.Subscript) -> None:
        if isinstance(node.value, ast.Name):
            self._update(node.value.id, node.ctx)
        self.generic_visit(node)

    def visit_For(self, node: ast.For) -> None:
        # Skip target
        self.visit(node.iter)
        for stmt in node.body:
            self.visit(stmt)
        for stmt in node.orelse:
            self.visit(stmt)


class ReadWrites(typing.NamedTuple):
    reads: dict[str, int]
    writes: dict[str, int]

    def __iter__(self) -> typing.Iterator[str]:
        return iter({**self.reads, **self.writes})

    @staticmethod
    def from_list(body: list[ast.AST] | list[ast.stmt]) -> ReadWrites:
        visitor = _ReadWriteVisitor()
        for node in body:
            visitor.visit(node)
        return visitor.rw

    @staticmethod
    def from_ast(node: ast.AST) -> ReadWrites:
        """
        Analyze an Abstract Syntax Tree (AST) node to determine the variables
        that are read and written within it.

        This function traverses the given AST node and collects information
        about variable reads and writes using the `_ReadWriteVisitor` class.

        :param node: The root AST node to analyze.
        :return: A `ReadWrites` object containing dictionaries of read and
                 written variable names.
        """
        visitor = _ReadWriteVisitor()
        visitor.visit(node)
        return visitor.rw


class _RenameVisitor(ast.NodeVisitor):
    def __init__(self, renames: dict[str, str]) -> None:
        super().__init__()
        self.renames = renames

    def visit_Name(self, node: ast.Name) -> None:
        node.id = self.renames.get(node.id, node.id)


def ast_rename(node: _A, renames: dict[str, str]) -> _A:
    """
    Rename variables in an Abstract Syntax Tree (AST) node, in-place.

    This function traverses the given AST node and renames variables
    based on the provided mapping of old names to new names.

    :param node: The root AST node to rename variables in.
    :param renames: A dictionary mapping old variable names to new variable names.
    :return: The modified AST node with variables renamed.
    """
    visitor = _RenameVisitor(renames)
    visitor.visit(node)
    return node


class _DeleteAssignments(ast.NodeTransformer):
    def __init__(self, to_remove: set[str]) -> None:
        super().__init__()
        self.to_remove = to_remove

    def visit_Assign(self, node: ast.Assign) -> ast.Assign | None:
        """
        Visit an assignment node and remove it if the target variable is in the to_remove set.

        :param node: The assignment node to visit.
        :return: The modified assignment node, or None if it should be removed.
        """
        if len(node.targets) == 1:
            (target,) = node.targets
            if isinstance(target, ast.Name) and target.id in self.to_remove:
                return None
        return node


def ast_delete_assignments(body: list[ast.AST], to_remove: set[str]) -> list[ast.AST]:
    new_body = []
    transformer = _DeleteAssignments(to_remove)
    for node in body:
        new_node = transformer.visit(node)
        if new_node is not None:
            new_body.append(new_node)
    return new_body



from __future__ import annotations

import random
from typing import TYPE_CHECKING

from .base_search import FlatConfig
from .base_search import PopulationBasedSearch
from .base_search import PopulationMember
from .base_search import performance
from .base_search import population_statistics

if TYPE_CHECKING:
    from collections.abc import Iterator
    from collections.abc import Sequence

    from ..runtime.config import Config
    from ..runtime.kernel import BoundKernel


class DifferentialEvolutionSearch(PopulationBasedSearch):
    """
    A search strategy that uses differential evolution to find the best config.
    """

    def __init__(
        self,
        # pyre-fixme[11]: BoundKernel undefined?
        kernel: BoundKernel,
        args: Sequence[object],
        population_size: int = 40,
        num_generations: int = 20,
        crossover_rate: float = 0.8,
        immediate_update: bool | None = None,
    ) -> None:
        super().__init__(kernel, args)
        if immediate_update is None:
            immediate_update = not kernel.settings.autotune_precompile
        self.population_size = population_size
        self.num_generations = num_generations
        self.crossover_rate = crossover_rate
        self.immediate_update = immediate_update

    def mutate(self, x_index: int) -> FlatConfig:
        a, b, c, *_ = [
            self.population[p]
            for p in random.sample(range(len(self.population)), 4)
            if p != x_index
        ]
        return self.config_gen.differential_mutation(
            self.population[x_index].flat_values,
            a.flat_values,
            b.flat_values,
            c.flat_values,
            self.crossover_rate,
        )

    def initial_two_generations(self) -> None:
        # The initial population is 2x larger so we can throw out the slowest half and give the tuning process a head start
        oversized_population = sorted(
            self.parallel_benchmark_flat(
                self.config_gen.random_population_flat(self.population_size * 2),
            ),
            key=performance,
        )
        self.log(
            "Initial population:",
            lambda: population_statistics(oversized_population),
        )
        self.population = oversized_population[: self.population_size]

    def iter_candidates(self) -> Iterator[tuple[int, PopulationMember]]:
        if self.immediate_update:
            for i in range(len(self.population)):
                yield i, self.benchmark_flat(self.mutate(i))
        else:
            yield from enumerate(
                self.parallel_benchmark_flat(
                    [self.mutate(i) for i in range(len(self.population))]
                )
            )

    def evolve_population(self) -> int:
        replaced = 0
        for i, candidate in self.iter_candidates():
            candidate = self.benchmark_flat(self.mutate(i))
            if candidate.perf < self.population[i].perf:
                self.population[i] = candidate
                replaced += 1
        return replaced

    def _autotune(self) -> Config:
        self.log(
            lambda: (
                f"Starting DifferentialEvolutionSearch with population={self.population_size}, "
                f"generations={self.num_generations}, crossover_rate={self.crossover_rate}"
            )
        )
        self.initial_two_generations()
        for i in range(2, self.num_generations):
            replaced = self.evolve_population()
            self.log(f"Generation {i}: replaced={replaced}", self.statistics)
        return self.best.config



from __future__ import annotations

import contextlib
import functools
from typing import Any
from typing import Callable
from typing import Generator

import torch
from torch._inductor.lowering import to_dtype
from torch._prims_common import ELEMENTWISE_TYPE_PROMOTION_KIND

inductor_lowering_dispatch: dict[  # pyre-ignore[5]
    Callable[..., Any] | str, Callable[..., Any]
] = {}


@contextlib.contextmanager
def patch_inductor_lowerings() -> Generator[  # pyre-ignore[3]
    None, Any, Any
]:
    """Context manager to temporarily patch the inductor lowering table.

    This is useful for overwriting specific Inductor lowerings without
    affecting the global state, especially in cases where Helion
    is missing support for a specific lowering.
    """
    original_lowerings = torch._inductor.lowering.lowerings.copy()
    try:
        torch._inductor.lowering.lowerings.update(inductor_lowering_dispatch)
        yield
    finally:
        torch._inductor.lowering.lowerings = original_lowerings


def _register_inductor_lowering(
    aten_fn: object,
    decomp_fn: object,
    broadcast: bool,
    type_promotion_kind: ELEMENTWISE_TYPE_PROMOTION_KIND | None,
    convert_input_to_bool: bool,
    lowering_dict: dict[object, Callable[..., object]],
) -> Callable[..., object]:
    from torch._inductor.lowering import fallbacks
    from torch._inductor.lowering import get_overloads
    from torch._inductor.lowering import in_namespace
    from torch._inductor.lowering import transform_args
    from torch._inductor.lowering import validate_ir

    @functools.wraps(decomp_fn)  # pyre-ignore[6]
    def wrapped(*args: Any, **kwargs: Any) -> object:
        args = list(args)  # pyre-ignore[9]
        kwargs = dict(kwargs)
        unpacked = False
        if len(args) == 1 and isinstance(args[0], (list, tuple)):
            unpacked = True
            args = list(args[0])  # pyre-ignore[9]

        if not all(
            (fn in fallbacks or in_namespace(fn, "_c10d_functional"))
            for fn in aten_fn  # pyre-ignore[16]
        ):
            # explicitly assert for "out=" ops for better error messages
            assert not any(x == "out" for x in kwargs), "out= ops aren't yet supported"

        args, kwargs = transform_args(  # pyre-ignore[9]
            args,  # pyre-ignore[6]
            kwargs,
            broadcast,
            type_promotion_kind,
            convert_input_to_bool,
        )

        if unpacked:
            args = [args]  # pyre-ignore[9]

        out = decomp_fn(*args, **kwargs)  # pyre-ignore[29]
        validate_ir(out)

        return out

    aten_fn = get_overloads(aten_fn)

    lowering_dict.update(dict.fromkeys(aten_fn, wrapped))
    return wrapped


# TODO(yf225): Switch to use upstream torch._inductor.lowering.register_lowering() after PyTorch 2.8 is released.
def register_inductor_lowering(
    aten_fn: object,
    broadcast: bool = False,
    type_promotion_kind: ELEMENTWISE_TYPE_PROMOTION_KIND
    | None = ELEMENTWISE_TYPE_PROMOTION_KIND.DEFAULT,
    convert_input_to_bool: bool = False,
    lowering_dict: dict[  # pyre-ignore[2]
        Any, Callable[..., Any]
    ] = inductor_lowering_dispatch,
) -> Callable[..., object]:
    return functools.partial(
        _register_inductor_lowering,
        aten_fn,
        broadcast=broadcast,
        type_promotion_kind=type_promotion_kind,
        convert_input_to_bool=convert_input_to_bool,
        lowering_dict=lowering_dict,
    )


def var_mean_helper_(
    x: torch._inductor.ir.TensorBox,
    *,
    axis: list[int] | None,
    correction: float | None,
    keepdim: bool,
    return_mean: bool,
) -> torch._inductor.ir.TensorBox:
    from torch._inductor.lowering import var_mean_sum_
    from torch._prims_common import get_computation_dtype

    out_dtype = x.get_dtype()
    compute_dtype = get_computation_dtype(out_dtype)
    x = to_dtype(x, compute_dtype, copy=False)

    kwargs = {
        "x": x,
        "axis": axis,
        "correction": correction,
        "keepdim": keepdim,
        "return_mean": return_mean,
    }
    # TODO(yf225): support Welford reduction in Helion, then switch back to use Inductor `var_mean_helper_()`.
    output = var_mean_sum_(**kwargs)
    output = tuple(to_dtype(o, out_dtype, copy=False) for o in output)
    return output[0] if not return_mean else output


@register_inductor_lowering(
    [torch.ops.aten.var.correction], lowering_dict=inductor_lowering_dispatch
)
def var_(
    x: torch._inductor.ir.TensorBox,
    axis: list[int] | None = None,
    *,
    correction: float | None = None,
    keepdim: bool = False,
) -> torch._inductor.ir.TensorBox:
    return var_mean_helper_(
        x,
        axis=axis,
        correction=correction,
        keepdim=keepdim,
        return_mean=False,
    )


@register_inductor_lowering(  # pyre-ignore[56]
    torch.ops.aten.var_mean.correction, lowering_dict=inductor_lowering_dispatch
)
def var_mean(
    x: torch._inductor.ir.TensorBox,
    axis: list[int] | None = None,
    *,
    correction: float | None = None,
    keepdim: bool = False,
) -> torch._inductor.ir.TensorBox:
    return var_mean_helper_(
        x,
        axis=axis,
        correction=correction,
        keepdim=keepdim,
        return_mean=True,
    )



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING

from torch._dynamo.source import AttrSource
from torch._dynamo.source import GetItemSource
from torch._dynamo.source import GlobalSource
from torch._dynamo.source import LocalSource

if TYPE_CHECKING:
    from torch._guards import Source

    from .host_function import HostFunction
    from .source_location import SourceLocation


@dataclasses.dataclass
class Origin:
    """Keeps track of where a variable came from."""

    def is_host(self) -> bool:
        """
        Check if the origin is a host.
        """
        return issubclass(self.base_type(), HostOrigin)

    def is_global(self) -> bool:
        """
        Check if the origin is a global variable.

        :return: True if the origin is from a global variable, False otherwise.
        """
        return issubclass(self.base_type(), GlobalOrigin)

    def is_argument(self) -> bool:
        """
        Check if the origin is an argument.

        :return: True if the origin is from an argument, False otherwise.
        """
        return issubclass(self.base_type(), ArgumentOrigin)

    def is_device(self) -> bool:
        """
        Check if the origin is a device.

        :return: True if the origin is a device, False otherwise.
        """
        return not self.is_host()

    def base_type(self) -> type[Origin]:
        """
        Get the base type of the origin, unwrapping things like attributes.

        :return: The base type of the origin.
        """
        return type(self)

    def needs_rename(self) -> bool:
        """
        Check if the origin needs to be renamed (globals and closures).

        :return: True if the origin needs to be renamed, False otherwise.
        """
        return self.is_global()

    def depth(self) -> int:
        """
        Get the depth of the origin.

        :return: The depth of the origin, which is 1 by default and increases each wrapper.
        """
        return 1

    def host_str(self) -> str:
        """
        Get a string representation of the host origin.

        :raises NotImplementedError: Always raises this error as it should be implemented by subclasses.
        """
        raise NotImplementedError(type(self).__name__)

    def suggest_var_name(self) -> str:
        """
        Suggest a variable name based on the origin.

        :raises NotImplementedError: Always raises this error as it should be implemented by subclasses.
        """
        raise NotImplementedError(type(self).__name__)

    def to_source(self) -> Source:
        """Convert to a PyTorch source object."""
        raise NotImplementedError(type(self).__name__)


@dataclasses.dataclass
class HostOrigin(Origin):
    pass


@dataclasses.dataclass
class NameOrigin(HostOrigin):
    """A variable that came from an ast.Name node."""

    name: str

    def __init__(self, name: str, function: HostFunction | None = None) -> None:
        super().__init__()
        self.name = name

    def host_str(self) -> str:
        return self.name

    def suggest_var_name(self) -> str:
        return self.name


class BuiltinOrigin(NameOrigin):
    def to_source(self) -> Source:
        return GlobalSource(self.name)


class GlobalOrigin(NameOrigin):
    def to_source(self) -> Source:
        return GlobalSource(self.name)


class ArgumentOrigin(NameOrigin):
    def to_source(self) -> Source:
        return LocalSource(self.name, is_input=True)


@dataclasses.dataclass
class WrappedOrigin(Origin):
    """Keeps track of where a variable came from."""

    value: Origin

    def base_type(self) -> type[Origin]:
        return self.value.base_type()

    def needs_rename(self) -> bool:
        return self.value.needs_rename()

    def depth(self) -> int:
        return 1 + self.value.depth()


@dataclasses.dataclass
class AttributeOrigin(WrappedOrigin):
    key: str

    def host_str(self) -> str:
        return f"{self.value.host_str()}.{self.key}"

    def suggest_var_name(self) -> str:
        return f"{self.value.suggest_var_name()}_attr_{self.key}"

    def to_source(self) -> Source:
        return AttrSource(self.value.to_source(), self.key)


@dataclasses.dataclass
class GetItemOrigin(WrappedOrigin):
    key: int | str

    def host_str(self) -> str:
        return f"{self.value.host_str()}[{self.key!r}]"

    def suggest_var_name(self) -> str:
        return f"{self.value.suggest_var_name()}_item_{self.key}"

    def to_source(self) -> Source:
        return GetItemSource(self.value.to_source(), self.key)


@dataclasses.dataclass
class TensorSizeOrigin(WrappedOrigin):
    key: int

    def host_str(self) -> str:
        return f"{self.value.host_str()}.size({self.key!r})"

    def suggest_var_name(self) -> str:
        return f"{self.value.suggest_var_name()}_size_{self.key}"

    def to_source(self) -> Source:
        return GetItemSource(AttrSource(self.value.to_source(), "shape"), self.key)


@dataclasses.dataclass
class ClosureOrigin(WrappedOrigin):
    key: int

    def needs_rename(self) -> bool:
        return True

    def host_str(self) -> str:
        return f"{self.value.host_str()}.__closure__[{self.key!r}].cell_contents"

    def suggest_var_name(self) -> str:
        return f"{self.value.suggest_var_name()}_closure_{self.key}"

    def to_source(self) -> Source:
        return AttrSource(
            GetItemSource(AttrSource(self.value.to_source(), "__closure__"), self.key),
            "cell_contents",
        )


@dataclasses.dataclass
class SourceOrigin(HostOrigin):
    location: SourceLocation


@dataclasses.dataclass
class DeviceOrigin(Origin):
    location: SourceLocation


@dataclasses.dataclass
class BlockSizeOrigin(Origin):
    block_id: int

    def host_str(self) -> str:
        """
        Get the host-side string representation of a block size variable.
        If the block size variable was not created (e.g., block size == 1),
        return the literal '1'.
        """
        from .device_function import DeviceFunction

        # Look up the block size variable name; if not set (e.g., size==1), use literal 1
        var = DeviceFunction.current().block_size_var(self.block_id)
        if var is None:
            return "1"
        return var

    def suggest_var_name(self) -> str:
        return f"block_size_{self.block_id}"


@dataclasses.dataclass
class ReductionDimensionOrigin(Origin):
    rdim_idx: int

    def host_str(self) -> str:
        raise NotImplementedError



from __future__ import annotations

import ast
import enum
import threading
import typing
from typing import TYPE_CHECKING
from typing import TypeVar

from .. import exc
from .source_location import SourceLocation
from .source_location import current_location

if TYPE_CHECKING:
    from collections.abc import Sequence

    from .type_propagation import TypeInfo

    _T = TypeVar("_T", bound=ast.AST)
    _R = TypeVar("_R")

    class _TLS(typing.Protocol):
        active_nodes: list[ExtendedAST]


tls: _TLS = typing.cast("_TLS", threading.local())


class LoopType(enum.Enum):
    UNSET = enum.auto()
    HOST = enum.auto()
    GRID = enum.auto()
    DEVICE = enum.auto()


class ExtendedAST:
    """
    We add some extra functionality to the AST classes, by dynamically
    subclassing each AST node class and mixing in this one.
    """

    # pyre-ignore[13]
    _fields: tuple[str, ...]

    def __init__(
        self,
        *,
        _location: SourceLocation,
        _type_info: TypeInfo | None = None,
        _loop_type: LoopType = LoopType.UNSET,
        _is_kernel_call: bool = False,
        _root_id: int | None = None,
        **kwargs: object,
    ) -> None:
        super().__init__(**kwargs)
        self._type_info: TypeInfo | None = _type_info
        self._location: SourceLocation = _location
        self._loop_type: LoopType = _loop_type
        self._is_kernel_call: bool = _is_kernel_call
        self._root_id: int | None = _root_id

    def new(self, fields: dict[str, object]) -> ExtendedAST:
        result = self.__class__(
            **fields,
            _location=self._location,
            _type_info=self._type_info,
            _loop_type=self._loop_type,
            _is_kernel_call=self._is_kernel_call,
            _root_id=self._root_id,
        )
        return self._location.to_ast(result)

    def fields(self) -> dict[str, object]:
        return {field: getattr(self, field) for field in self._fields}

    def copy(self, **changes: object) -> ExtendedAST:
        return self.new({**self.fields(), **changes})

    def __repr__(self) -> str:
        assert isinstance(self, ast.AST)
        return ast.dump(self)

    def update_type_info(self, type_info: TypeInfo) -> TypeInfo:
        if self._type_info is not None and type_info != self._type_info:
            type_info = self._type_info.merge(type_info)
        self._type_info = type_info
        return self._type_info

    def debug_annotations(self) -> list[str]:
        result = []
        if self._type_info:
            result.extend(self._type_info.debug_annotations())
        if self._loop_type != LoopType.UNSET:
            result.append(f"loop_type={self._loop_type.name}")
        return result

    def __enter__(self) -> None:
        try:
            tls.active_nodes.append(self)
        except AttributeError:
            tls.active_nodes = [self]
        self._location.__enter__()

    def __exit__(self, *args: object) -> None:
        self._location.__exit__(*args)
        tls.active_nodes.pop()

    @staticmethod
    def current() -> Sequence[ExtendedAST]:
        """Stack of nodes currently being processed."""
        try:
            return tls.active_nodes
        except AttributeError:
            tls.active_nodes = rv = []
            return rv


_to_extended: dict[type[ast.AST], type[ast.AST]] = {}


def get_wrapper_cls(cls: type[ast.AST]) -> type[ast.AST]:
    if new_cls := _to_extended.get(cls):
        return new_cls

    class Wrapper(ExtendedAST, cls):
        pass

    Wrapper.__name__ = cls.__name__
    rv = typing.cast("type[ast.AST]", Wrapper)
    _to_extended[cls] = rv
    return rv


def create(cls: type[_T], **fields: object) -> _T:
    # pyre-ignore[28]
    result = get_wrapper_cls(cls)(**fields, _location=current_location())
    assert isinstance(result, ExtendedAST)
    result._location.to_ast(result)
    return typing.cast("_T", result)


def create_arg(name: str, annotation: str | None = None) -> ast.arg:
    return create(
        ast.arg,
        arg=name,
        annotation=expr_from_string(annotation) if annotation else None,
        type_comment=None,
    )


def create_arguments(args: list[ast.arg]) -> ast.arguments:
    return create(
        ast.arguments,
        args=args,
        posonlyargs=[],
        defaults=[],
        kw_defaults=[],
        kwonlyargs=[],
    )


def statement_from_string(template: str, **placeholders: ast.AST) -> ast.stmt:
    (statement,) = ast.parse(template).body
    location: SourceLocation = current_location()

    def _replace(node: _R) -> _R:
        if isinstance(node, list):
            return [_replace(item) for item in node]
        if not isinstance(node, ast.AST):
            return node
        if isinstance(node, ast.Name) and node.id in placeholders:
            return placeholders[node.id]
        cls = get_wrapper_cls(type(node))
        return location.to_ast(
            cls(
                **{field: _replace(getattr(node, field)) for field in node._fields},
                _location=location,
            )
        )

    return _replace(statement)


def expr_from_string(template: str, **placeholders: ast.AST) -> ast.AST:
    expr = statement_from_string(template, **placeholders)
    assert isinstance(expr, ast.Expr)
    return expr.value


def convert(node: ast.AST) -> ast.AST:
    if isinstance(node, ast.AST):
        cls = get_wrapper_cls(type(node))
        if "lineno" in node._attributes:
            location = SourceLocation.from_ast(node)
        else:
            # some nodes like arguments lack location information
            location = current_location()
        with location:
            # pyre-ignore[28]
            return cls(
                **{field: convert(getattr(node, field)) for field in node._fields},
                **{attr: getattr(node, attr) for attr in node._attributes},
                _location=location,
            )
    elif isinstance(node, list):
        return [convert(item) for item in node]
    else:
        return node


class NodeVisitor(ast.NodeVisitor):
    def visit(self, node: ast.AST) -> ast.AST:
        assert isinstance(node, ExtendedAST)
        with node:
            try:
                visitor = getattr(
                    self,
                    f"visit_{node.__class__.__name__}",
                    self.generic_visit,
                )
                return visitor(node)
            except exc.Base:
                raise
            except Exception as e:
                raise exc.InternalError(e) from e


# Determine whether vanilla ast.unparse keeps parentheses in "(a, b) = c".
# If so, remove the parentheses via `_TupleParensRemovedUnparser` below.
# NOTE: this is to make Python source format consistent between Python 3.10 and 3.12+
_test_src: str = "(a, b) = c"
_needs_to_remove_tuple_parens: bool = (
    ast.unparse(ast.parse(_test_src)).lstrip().startswith("(")
)


class _TupleParensRemovedUnparser(ast._Unparser):  # pyre-ignore[11]
    def visit_Tuple(self, node) -> None:  # pyre-ignore[2]
        if _needs_to_remove_tuple_parens and isinstance(
            getattr(node, "ctx", None), ast.Store
        ):
            if len(node.elts) == 1:  # single-element tuple
                self.traverse(node.elts[0])  # pyre-ignore[16]
                self.write(",")  # pyre-ignore[16]
            else:  # multi-element tuple
                self.interleave(  # pyre-ignore[16]
                    lambda: self.write(", "), self.traverse, node.elts
                )
            return
        # For everything else fall back to default behavior
        super().visit_Tuple(node)  # pyre-ignore[16]


def unparse(ast_obj: ast.AST) -> str:
    unparser = _TupleParensRemovedUnparser()
    return unparser.visit(ast_obj)  # pyre-ignore[16]



from __future__ import annotations

from .constexpr import ConstExpr as constexpr  # noqa: F401
from .constexpr import specialize as specialize
from .creation_ops import full as full
from .creation_ops import zeros as zeros
from .device_print import device_print as device_print
from .loops import grid as grid
from .loops import tile as tile
from .memory_ops import atomic_add as atomic_add
from .memory_ops import load as load
from .memory_ops import store as store
from .tile_ops import tile_begin as tile_begin
from .tile_ops import tile_block_size as tile_block_size
from .tile_ops import tile_end as tile_end
from .tile_ops import tile_id as tile_id
from .tile_ops import tile_index as tile_index
from .tile_proxy import Tile as Tile
from .tunable_ops import register_block_size as register_block_size
from .tunable_ops import register_reduction_dim as register_reduction_dim
from .tunable_ops import register_tunable as register_tunable
from .view_ops import subscript as subscript



from __future__ import annotations

import ast
import collections
import dataclasses
import functools
import itertools
import operator
from typing import TYPE_CHECKING
from typing import NamedTuple
from typing import TypeVar
import weakref

import sympy
import torch

from .ast_extension import create
from .ast_extension import expr_from_string
from .ast_extension import statement_from_string
from .compile_environment import CompileEnvironment
from .compile_environment import _to_sympy
from .host_function import HostFunction
from .program_id import GridProgramIDs
from .program_id import L2GroupingProgramIDs
from .program_id import ProgramID
from .program_id import ProgramIDs
from .program_id import SharedProgramID
from .program_id import VirtualProgramIDs

if TYPE_CHECKING:
    from collections.abc import Sequence

    from .device_function import DeviceFunction
    from .inductor_lowering import CodegenState

    _T = TypeVar("_T")
    SymIntLike = torch.SymInt | int
    ShapeLike = Sequence[SymIntLike]


@dataclasses.dataclass
class DeviceLoopOrGridState:
    strategy: TileStrategy
    end_var_name: dict[int, str]

    @property
    def block_ids(self) -> list[int]:
        return self.strategy.block_ids


@dataclasses.dataclass
class DeviceLoopState(DeviceLoopOrGridState):
    for_node: ast.For
    inner_statements: list[ast.AST]
    end_bounds: dict[int, sympy.Expr | None]
    outer_prefix: list[ast.AST] = dataclasses.field(default_factory=list)
    outer_suffix: list[ast.AST] = dataclasses.field(default_factory=list)


class DeviceGridState(DeviceLoopOrGridState):
    pass


class PersistentReductionState(DeviceLoopOrGridState):
    pass


class TileStrategy:
    _fn: weakref.ReferenceType[DeviceFunction]
    block_ids: list[int]

    def __init__(
        self,
        fn: DeviceFunction,
        block_ids: list[int],
    ) -> None:
        self._fn = weakref.ref(fn)
        self.block_ids = block_ids
        self.index_vars: dict[int, str] = {
            block_idx: self.fn.new_var(f"indices_{block_idx}", dce=True)
            for block_idx in block_ids
        }
        self.offset_vars: dict[int, str] = {
            block_idx: self.fn.new_var(f"offset_{block_idx}", dce=True)
            for block_idx in block_ids
        }

    @property
    def fn(self) -> DeviceFunction:
        fn = self._fn()
        assert fn is not None
        return fn

    def offset_var(self, block_idx: int) -> str:
        return self.offset_vars[block_idx]

    def index_var(self, block_idx: int) -> str:
        return self.index_vars[block_idx]

    def mask_var(self, block_idx: int) -> str | None:
        raise NotImplementedError

    def block_size_var(self, block_idx: int) -> str | None:
        return self.fn.block_size_var_cache.get((block_idx,))

    def user_size(self, block_index: int) -> sympy.Expr:
        raise NotImplementedError

    def codegen_grid(self, state: CodegenState) -> DeviceGridState:
        raise NotImplementedError

    def codegen_device_loop(self, state: CodegenState) -> DeviceLoopState:
        raise NotImplementedError

    def codegen_preamble(self, state: CodegenState) -> None:
        """Called after a *different* strategy has been used to generate the grid."""

    def compact_shape(self, shapes: list[CompactedShape]) -> list[CompactedShape]:
        raise NotImplementedError


class BlockSizeTileStrategy(TileStrategy):
    def __init__(
        self,
        fn: DeviceFunction,
        block_ids: list[int],
        block_size: list[SymIntLike] | SymIntLike,
        loop_order: list[int],
    ) -> None:
        super().__init__(
            fn=fn,
            block_ids=block_ids,
        )
        self.block_size = block_size
        self.loop_order = loop_order

    def _reorder(self, block_ids: list[_T]) -> list[_T]:
        if len(block_ids) <= 1:
            return block_ids
        order = self.loop_order
        assert len(order) == len(block_ids), (
            f"Invalid order length: {len(order)} != {len(block_ids)}"
        )
        assert {*order} == {*range(len(order))}, f"Invalid permutation: {order}"
        return [block_ids[i] for i in reversed(order)]

    def user_size(self, block_index: int) -> sympy.Expr:
        return CompileEnvironment.current().block_sizes[block_index].symbol()

    def get_end_bounds(self, state: CodegenState) -> dict[int, sympy.Expr | None]:
        block_ids = self.block_ids
        _, _, ends, _ = state.proxy_args
        assert isinstance(ends, list)
        bounds = {}
        for block_idx, end in zip(block_ids, ends, strict=True):
            if isinstance(end, (int, torch.SymInt)):
                end = _to_sympy(end)
            else:
                end = None
            bounds[block_idx] = end
        return bounds


class FlattenedTileStrategy(BlockSizeTileStrategy):
    """Collapse all dimensions into single flat iteration space."""

    block_size: SymIntLike

    def __init__(
        self,
        fn: DeviceFunction,
        block_ids: list[int],
        block_size: list[SymIntLike] | SymIntLike,
        loop_order: list[int],
    ) -> None:
        assert isinstance(block_size, (int, torch.SymInt))
        super().__init__(fn, block_ids, block_size, loop_order)
        env = CompileEnvironment.current()
        if env.known_multiple(
            functools.reduce(
                operator.mul, [env.block_sizes[i].numel for i in block_ids]
            ),
            block_size,
        ):
            self._mask_var: str | None = None
        else:
            self._mask_var = self.new_var("mask", dce=True)

        key = (*self.block_ids,)
        assert key not in fn.block_size_var_cache
        fn.block_size_var_cache[key] = bs_var = self.new_var("_BLOCK_SIZE")
        for block_index in block_ids:
            fn.block_size_var_cache[(block_index,)] = bs_var

    def new_var(self, prefix: str, dce: bool = False) -> str:
        return self.fn.new_var(
            f"{prefix}_{'_'.join(map(str, self.block_ids))}", dce=dce
        )

    def offset_var(self, block_idx: int) -> str:
        raise NotImplementedError("offset_var not used in FlattenedTileStrategy")

    def mask_var(self, block_idx: int) -> str | None:
        return self._mask_var

    def block_size_var(self, block_idx: int) -> str:
        return self.fn.block_size_var_cache[tuple(self.block_ids)]

    def _codegen_common(
        self, state: CodegenState
    ) -> tuple[str, str, sympy.Expr, list[ast.AST]]:
        block_ids = self.block_ids
        env = CompileEnvironment.current()
        total_numel = sympy.S.One
        offsets_var = self.new_var("offsets", dce=True)
        block_size_var = self.block_size_var(-1)
        statements = []
        if state.device_function.constexpr_arg(block_size_var):
            block_size_str = HostFunction.current().literal_expr(self.block_size)
            state.codegen.host_statements.append(
                statement_from_string(f"{block_size_var} = {block_size_str}")
            )
        for i, block_idx in enumerate(self._reorder(block_ids)):
            # need to get the block size
            numel = env.block_sizes[block_idx].numel
            block_index_var = self.index_var(block_idx)
            expr = offsets_var
            if total_numel != sympy.S.One:
                expr = f"({expr}) // ({state.sympy_expr(total_numel)})"
            if i + 1 < len(block_ids):
                expr = f"({expr}) % ({state.sympy_expr(numel)})"
            statements.append(statement_from_string(f"{block_index_var} = {expr}"))
            total_numel = total_numel * numel

        mask_var = self.mask_var(-1)
        if mask_var is not None:
            statements.append(
                statement_from_string(
                    f"{mask_var} = {offsets_var} < ({state.sympy_expr(total_numel)})"
                )
            )
        return block_size_var, offsets_var, total_numel, statements

    def codegen_grid(self, state: CodegenState) -> DeviceGridState:
        block_size_var, offsets_var, total_numel, statements = self._codegen_common(
            state
        )
        env = CompileEnvironment.current()
        dtype = env.triton_index_type()
        state.add_statement(
            f"{offsets_var} = tl.program_id(0) * ({block_size_var}) + tl.arange(0, {block_size_var}).to({dtype})"
        )
        state.codegen.statements_stack[-1].extend(statements)

        class TmpPid(ProgramIDs):
            def codegen_grid(self) -> ast.AST:
                return expr_from_string(
                    f"(triton.cdiv({HostFunction.current().sympy_expr(total_numel)}, {block_size_var}), 1, 1)"
                )

        state.device_function.set_pid(TmpPid())

        end_var_name = {}
        for block_id in self.block_ids:
            end_bound = env.block_sizes[block_id].numel
            end_var_name[block_id] = state.sympy_expr(end_bound)
        return DeviceGridState(self, end_var_name=end_var_name)

    def codegen_device_loop(self, state: CodegenState) -> DeviceLoopState:
        block_size_var, offsets_var, total_numel, statements = self._codegen_common(
            state
        )
        dtype = CompileEnvironment.current().triton_index_type()
        lid = self.new_var("lid")
        for_node = create(
            ast.For,
            target=create(ast.Name, id=lid, ctx=ast.Store()),
            iter=expr_from_string(
                f"range(tl.cdiv({state.sympy_expr(total_numel)}, {block_size_var}))"
            ),
            body=(
                body := [
                    statement_from_string(
                        f"{offsets_var} = {lid} * {block_size_var} + tl.arange(0, {block_size_var}).to({dtype})"
                    ),
                    *statements,
                ]
            ),
            orelse=[],
            type_comment=None,
        )
        return DeviceLoopState(
            self,
            for_node=for_node,
            inner_statements=body,
            end_bounds=self.get_end_bounds(state),
            end_var_name={},
        )

    @classmethod
    def update_allow_flattened(cls, shape: Sequence[sympy.Expr]) -> None:
        env = CompileEnvironment.current()
        used_indices = {}
        for i, x in enumerate(shape):
            block_idx = env.get_block_id(x)
            if block_idx is not None:
                used_indices[block_idx] = i
        flatten_loops = env.config_spec.flatten_loops
        for spec in [*flatten_loops]:
            block_ids = spec.block_ids
            if not (
                all(x in used_indices for x in block_ids)
                or all(x not in used_indices for x in block_ids)
            ):
                flatten_loops.disable_block_id(block_ids[0])
                continue
            for i, j in itertools.pairwise(block_ids):
                if i in used_indices and used_indices[i] + 1 != used_indices[j]:
                    # The block indices must be contiguous
                    flatten_loops.disable_block_id(block_ids[0])
                    break

    def compact_shape(self, shapes: list[CompactedShape]) -> list[CompactedShape]:
        output = []
        shape_queue = collections.deque(shapes)
        while shape_queue:
            shape = shape_queue.popleft()
            if len(shape.block_ids) != 1 or shape.block_ids[0] not in self.block_ids:
                output.append(shape)
                continue
            assert shape.block_ids[0] == self.block_ids[0]
            for expected in self.block_ids[1:]:
                new_shape = shape_queue.popleft()
                assert len(new_shape.block_ids) == 1
                assert new_shape.block_ids[0] == expected
                shape = shape.combine(new_shape)
            output.append(shape)
        return output


class _BaseNDTileStrategy(BlockSizeTileStrategy):
    block_size: list[SymIntLike]

    def __init__(
        self,
        fn: DeviceFunction,
        block_ids: list[int],
        block_size: list[SymIntLike] | SymIntLike,
        loop_order: list[int],
    ) -> None:
        assert isinstance(block_size, list)
        super().__init__(fn, block_ids, block_size, loop_order)
        for bs, block_idx in zip(block_size, block_ids, strict=True):
            if (block_idx,) not in fn.block_size_var_cache and bs != 1:
                fn.block_size_var_cache[(block_idx,)] = fn.new_var(
                    f"_BLOCK_SIZE_{block_idx}"
                )

    def codegen_grid(self, state: CodegenState) -> DeviceGridState:
        block_ids = self.block_ids
        env = CompileEnvironment.current()
        device_function = state.device_function
        dtype = env.triton_index_type()
        block_sizes = self.block_size
        assert len(block_sizes) == len(block_ids)
        if isinstance(state.device_function.pid, SharedProgramID):
            # Disable for shared pid
            self.fn.config.config["use_yz_grid"] = False
        pids = self.select_pid_strategy()
        if isinstance(state.device_function.pid, SharedProgramID):
            pids.shared_pid_var = state.device_function.pid.shared_pid_var
        for i, (block_idx, block_size) in enumerate(
            reversed(self._reorder([*zip(block_ids, block_sizes, strict=True)]))
        ):
            numel = env.block_sizes[block_idx].numel
            offset_var = self.offset_var(block_idx)
            index_var = self.index_var(block_idx)
            pid_var = device_function.new_var(f"pid_{i}", dce=True)
            if block_size != 1:
                block_size_var = self.block_size_var(block_idx)
                assert block_size_var is not None
                # TODO(jansel): need to check for conflict with user variable names since block_size_var is on host
                if state.device_function.constexpr_arg(block_size_var):
                    state.codegen.host_statements.append(
                        statement_from_string(
                            f"{block_size_var} = {HostFunction.current().literal_expr(block_size)}"
                        )
                    )
                state.add_statement(f"{offset_var} = {pid_var} * {block_size_var}")
                state.add_statement(
                    f"{index_var} = ({offset_var} + tl.arange(0, ({block_size_var}))).to({dtype})"
                )
            else:
                block_size_var = "1"
                dtype = env.triton_index_type()
                state.add_statement(f"{offset_var} = {pid_var}")
                state.add_statement(
                    f"{index_var} = {offset_var} + tl.zeros([1], {dtype})"
                )
            mask_statement = self._setup_mask(  # pyre-ignore[16]
                state, block_idx, block_size, index_var, numel
            )
            if mask_statement is not None:
                state.add_statement(mask_statement)
            pid = ProgramID(pid_var, block_size_var, numel)
            pids.append(pid)
        pids.codegen(state)
        if isinstance(state.device_function.pid, SharedProgramID):
            shared_pid = state.device_function.pid
            shared_pid.pids.append(pids)
            shared_pid.codegen(state)
        else:
            state.device_function.set_pid(pids)

        # Extract end_var_name from end bound expressions
        end_var_name = {}
        for block_id in self.block_ids:
            end_bound = env.block_sizes[block_id].numel
            end_var_name[block_id] = state.sympy_expr(end_bound)
        return DeviceGridState(self, end_var_name=end_var_name)

    def select_pid_strategy(self) -> ProgramIDs:
        if 1 < len(self.block_ids) <= 3 and self.fn.config.use_yz_grid:
            return GridProgramIDs()
        return VirtualProgramIDs()

    def _to_ast(self, x: object, to_dtype: str | None = None) -> ast.AST:
        if isinstance(x, ast.AST):
            if to_dtype:
                return expr_from_string(f"value.to({to_dtype})", value=x)
            return x
        if isinstance(x, int):
            return expr_from_string(repr(x))
        if isinstance(x, sympy.Expr):
            from .device_function import DeviceFunction

            return expr_from_string(DeviceFunction.current().sympy_expr(x))
        raise NotImplementedError(f"{type(x)} is not implemented.")

    def codegen_device_loop(self, state: CodegenState) -> DeviceLoopState:
        # TODO(jansel): refactor this to share code with codegen_grid
        block_ids = self.block_ids
        env = CompileEnvironment.current()
        dtype = env.triton_index_type()
        block_sizes = self.block_size
        body = innermost_body = []
        for_node: ast.For | None = None
        assert len(block_sizes) == len(block_ids)
        _, begins, ends, _ = state.ast_args
        assert isinstance(begins, list)
        assert isinstance(ends, list)
        end_var_name = {}
        for block_idx, block_size, begin, end in self._reorder(
            [*zip(block_ids, block_sizes, begins, ends, strict=True)]
        ):
            offset_var = self.offset_var(block_idx)
            index_var = self.index_var(block_idx)
            if block_size != 1:
                block_size_var = self.block_size_var(block_idx)
                assert block_size_var is not None
                if state.device_function.constexpr_arg(block_size_var):
                    state.codegen.host_statements.append(
                        statement_from_string(
                            f"{block_size_var} = {HostFunction.current().literal_expr(block_size)}"
                        )
                    )
            else:
                block_size_var = "1"
            end_var_name[block_idx] = state.codegen.lift(
                self._to_ast(end, to_dtype=dtype), dce=True, prefix="end"
            ).id
            for_node = create(
                ast.For,
                target=create(ast.Name, id=offset_var, ctx=ast.Store()),
                iter=expr_from_string(
                    f"range(begin, end, {block_size_var})",
                    begin=self._to_ast(begin, to_dtype=dtype),
                    end=self._to_ast(end, to_dtype=dtype),
                ),
                body=body,
                orelse=[],
                type_comment=None,
            )
            assert for_node.body is body
            extra_body = [
                statement_from_string(
                    f"{index_var} = {offset_var} + tl.arange(0, ({block_size_var})).to({dtype})"
                ),
            ]
            mask_statement = self._setup_mask(  # pyre-ignore[16]
                state, block_idx, block_size, index_var, end
            )
            if mask_statement is not None:
                extra_body.append(mask_statement)
            body[:] = [*extra_body, *body]
            body = [for_node]
        assert for_node is not None
        return DeviceLoopState(
            self,
            for_node=for_node,
            inner_statements=innermost_body,
            end_bounds=self.get_end_bounds(state),
            end_var_name=end_var_name,
        )

    def compact_shape(self, shapes: list[CompactedShape]) -> list[CompactedShape]:
        # TODO(jansel): we should combine size==1 dimensions here
        return shapes


class NDTileStrategy(_BaseNDTileStrategy):
    """Do up to 3D tiling using the kernel grid."""

    def __init__(
        self,
        fn: DeviceFunction,
        block_ids: list[int],
        block_size: list[SymIntLike] | SymIntLike,
        loop_order: list[int],
        l2_grouping: int,
    ) -> None:
        super().__init__(fn, block_ids, block_size, loop_order)
        self.mask_vars: dict[int, str | None] = {}
        self.l2_grouping = l2_grouping

    def mask_var(self, block_idx: int) -> str | None:
        return self.mask_vars[block_idx]

    def _setup_mask(
        self,
        state: CodegenState,
        block_idx: int,
        block_size: SymIntLike,
        index_var: str,
        end: object,
    ) -> ast.stmt | None:
        if (
            CompileEnvironment.current()
            .block_sizes[block_idx]
            .known_multiple(block_size)
        ):
            self.mask_vars[block_idx] = None
            return None
        self.mask_vars[block_idx] = mask_var = self.fn.new_var(
            f"mask_{block_idx}", dce=True
        )
        return statement_from_string(
            f"{mask_var} = ({index_var}) < end", end=self._to_ast(end)
        )

    def select_pid_strategy(self) -> ProgramIDs:
        if self.l2_grouping > 1:
            return L2GroupingProgramIDs(group_size=self.l2_grouping)
        return super().select_pid_strategy()


class NDGridTileStrategy(_BaseNDTileStrategy):
    def __init__(
        self,
        fn: DeviceFunction,
        block_ids: list[int],
        loop_order: list[int],
    ) -> None:
        super().__init__(
            fn=fn,
            block_ids=block_ids,
            block_size=[1] * len(block_ids),  # pyre-ignore[6]
            loop_order=loop_order,
        )

    def mask_var(self, block_idx: int) -> str | None:
        return None

    def _setup_mask(
        self,
        *args: object,
        **kwargs: object,
    ) -> None:
        return None


class CompactedShape(NamedTuple):
    size_str: str
    user_indices: list[int]
    block_ids: list[int]

    def combine(self, other: CompactedShape) -> CompactedShape:
        size_str = self.size_str
        if size_str == "1":
            size_str = other.size_str
        else:
            assert other.size_str in ("1", size_str)
        return CompactedShape(
            size_str=size_str,
            user_indices=[*self.user_indices, *other.user_indices],
            block_ids=[*self.block_ids, *other.block_ids],
        )



from __future__ import annotations

import dataclasses
import functools
import operator
from typing import TYPE_CHECKING

from torch._inductor.runtime.runtime_utils import next_power_of_2

from .._compat import supports_tensor_descriptor
from ..exc import InvalidConfig
from .block_id_sequence import BlockIdSequence
from .block_id_sequence import _BlockIdItem
from .block_id_sequence import _PowerOfTwoBlockIdItem
from .config_fragment import BlockSizeFragment
from .config_fragment import BooleanFragment
from .config_fragment import ConfigSpecFragment
from .config_fragment import EnumFragment
from .config_fragment import IntegerFragment
from .config_fragment import NumWarpsFragment
from .config_fragment import PermutationFragment
from .config_fragment import PowerOfTwoFragment
from .config_fragment import assert_integer_power_of_two
import helion

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Sequence

DEFAULT_NUM_WARPS = 4
DEFAULT_NUM_STAGES = 3
VALID_KEYS: frozenset[str] = frozenset(
    [
        "block_sizes",
        "loop_orders",
        "l2_groupings",
        "reduction_loops",
        "flatten_loops",
        "num_warps",
        "num_stages",
        "use_yz_grid",
        "indexing",
    ]
)


@dataclasses.dataclass
class ConfigSpec:
    block_sizes: BlockIdSequence[BlockSizeSpec] = dataclasses.field(
        default_factory=BlockIdSequence
    )
    loop_orders: BlockIdSequence[LoopOrderSpec] = dataclasses.field(
        default_factory=BlockIdSequence
    )
    l2_groupings: BlockIdSequence[L2GroupingSpec] = dataclasses.field(
        default_factory=BlockIdSequence
    )
    flatten_loops: BlockIdSequence[FlattenLoopSpec] = dataclasses.field(
        default_factory=BlockIdSequence
    )
    reduction_loops: BlockIdSequence[ReductionLoopSpec] = dataclasses.field(
        default_factory=BlockIdSequence
    )
    user_defined_tunables: dict[str, ConfigSpecFragment] = dataclasses.field(
        default_factory=dict
    )
    allow_use_yz_grid: bool | None = None

    def _remove_duplicates(self) -> None:
        self.loop_orders._remove_duplicates()
        self.l2_groupings._remove_duplicates()
        self.flatten_loops._remove_duplicates()

    def normalize(self, config: helion.Config | dict[str, object]) -> None:
        """Normalize the config to match the block_sizes and validate the config."""
        if isinstance(config, helion.Config):
            self.normalize(config.config)
            return

        for name in (
            "block_size",
            "loop_order",
            "reduction_loop",
            "l2_grouping",
            "flatten_loop",
        ):
            if name in config:
                names = f"{name}s"
                if names in config:
                    raise InvalidConfig(f"Cannot specify both {name} and {names}")
                config[names] = [config.pop(name)]

        for name, mapping, flatten in [
            ("block_sizes", self.block_sizes, True),
            ("flatten_loops", self.flatten_loops, True),
            ("l2_groupings", self.l2_groupings, True),
            ("loop_orders", self.loop_orders, False),
            ("reduction_loops", self.reduction_loops, True),
        ]:
            config[name] = mapping._normalize(
                name, config.get(name, ()), flatten=flatten
            )

        for name in ("loop_orders", "l2_groupings", "flatten_loops", "reduction_loops"):
            if not config[name]:
                config.pop(name)

        config.setdefault("num_warps", DEFAULT_NUM_WARPS)
        config.setdefault("num_stages", DEFAULT_NUM_STAGES)
        # TODO(jansel): include num_ctas and max_nreg

        if self.allow_use_yz_grid:
            config.setdefault("use_yz_grid", False)

        config.setdefault("indexing", "pointer")

        # Allow tunable parameter keys in addition to VALID_KEYS
        allowed_keys = VALID_KEYS | {*self.user_defined_tunables.keys()}
        if invalid_keys := ({*config} - allowed_keys):
            raise InvalidConfig(f"Invalid config keys {sorted(invalid_keys)!r}")

    def default_config(self) -> helion.Config:
        return self.flat_config(lambda x: x.default())

    def flat_config(self, fn: Callable[[ConfigSpecFragment], object]) -> helion.Config:
        """Map a flattened version of the config using the given function."""
        config = {
            "block_sizes": self.block_sizes._flat_config(self, fn),
            "loop_orders": self.loop_orders._flat_config(self, fn),
            "flatten_loops": self.flatten_loops._flat_config(self, fn),
            "l2_groupings": self.l2_groupings._flat_config(self, fn),
            "reduction_loops": self.reduction_loops._flat_config(self, fn),
            "num_warps": fn(NumWarpsFragment(1, 32, DEFAULT_NUM_WARPS)),
            "num_stages": fn(IntegerFragment(1, 8, DEFAULT_NUM_STAGES)),
            "indexing": fn(
                EnumFragment(
                    ("pointer", "block_ptr", "tensor_descriptor")
                    if supports_tensor_descriptor()
                    else ("pointer", "block_ptr")
                )
            ),
        }
        # Add tunable parameters
        for key, fragment in self.user_defined_tunables.items():
            config[key] = fn(fragment)

        if self.allow_use_yz_grid:
            use_yz_grid = fn(BooleanFragment())
            # pyre-ignore[16]
            if (not config["l2_groupings"] or config["l2_groupings"][0] == 1) and (
                not config["flatten_loops"] or not config["flatten_loops"][0]
            ):
                config["use_yz_grid"] = use_yz_grid
        for name in ("loop_orders", "flatten_loops", "reduction_loops", "l2_groupings"):
            if not config[name]:
                config.pop(name)
        return helion.Config(**config)  # pyre-ignore[6]


class LoopOrderSpec(_BlockIdItem):
    def _fragment(self, base: ConfigSpec) -> PermutationFragment:
        return PermutationFragment(len(self.block_ids))

    def _normalize(self, name: str, value: object) -> list[int]:
        if type(value) is not list:
            if not isinstance(value, tuple):
                raise InvalidConfig(f"{name} must be a list, got {value!r}")
            value = [*value]
        length = len(self.block_ids)
        if len(value) != length:
            raise InvalidConfig(f"{name} must be length {length}, got {len(value)}")
        if {*value} != {*range(length)}:
            raise InvalidConfig(f"{name} must be permutation, got {value!r}")
        return value

    def _fill_missing(self) -> list[int]:
        """Provide a value when not provided by the user."""
        return [*range(len(self.block_ids))]


class L2GroupingSpec(_PowerOfTwoBlockIdItem):
    def _fragment(self, base: ConfigSpec) -> PowerOfTwoFragment:
        return PowerOfTwoFragment(1, 64, 1)

    def _fill_missing(self) -> int:
        return 1


class BlockSizeSpec(_PowerOfTwoBlockIdItem):
    def __init__(
        self,
        *,
        block_id: int,
        size_hint: int,
        min_size: int = 1,
        max_size: int | None = None,
    ) -> None:
        super().__init__([block_id])
        self.size_hint = size_hint
        self.min_size: int = min_size
        self.max_size: int = (
            next_power_of_2(size_hint) if max_size is None else max_size
        )
        assert self.min_size <= self.max_size

    def __repr__(self) -> str:
        fields = []
        for field, default in (
            ("block_id", None),
            ("size_hint", None),
            ("min_size", 1),
            ("max_size", next_power_of_2(self.size_hint)),
        ):
            value = getattr(self, field)
            if value != default:
                fields.append(f"{field}={value!r}")
        return f"BlockSizeSpec({', '.join(fields)})"

    def update_min(self, value: int) -> None:
        self.min_size = assert_integer_power_of_two(max(value, self.min_size))
        if self.max_size < self.min_size:
            self.max_size = self.min_size

    def update_max(self, value: int) -> None:
        self.max_size = assert_integer_power_of_two(min(value, self.max_size))

    def update_hint(self, value: int) -> None:
        self.size_hint = value
        self.update_max(next_power_of_2(value))

    def _fragment(self, base: ConfigSpec) -> BlockSizeFragment:
        total_ndim = len(base.block_sizes)
        reduction_numel = _product(
            [next_power_of_2(spec.size_hint) for spec in base.reduction_loops]
        )
        if total_ndim <= 1 and reduction_numel <= 1:
            default = 1024
        elif total_ndim <= 2 and reduction_numel <= 128:
            default = 32
        elif reduction_numel <= 256:
            default = 16
        else:
            default = 1
        return BlockSizeFragment(
            self.min_size,
            self.max_size,
            default,
        )


class FlattenLoopSpec(_BlockIdItem):
    def _fragment(self, base: ConfigSpec) -> BooleanFragment:
        return BooleanFragment()

    def _normalize(self, name: str, value: object) -> bool:
        if not isinstance(value, bool):
            raise InvalidConfig(f"{name} must be a boolean, got {value!r}") from None
        return value

    def _fill_missing(self) -> bool:
        return False


class ReductionLoopSpec(_PowerOfTwoBlockIdItem):
    def __init__(
        self,
        *,
        block_id: int,
        size_hint: int,
    ) -> None:
        super().__init__([block_id])
        self.size_hint = size_hint

    def _flat_config(
        self, base: ConfigSpec, fn: Callable[[ConfigSpecFragment], object]
    ) -> int | None:
        low = 8  # TODO(jansel): is smaller needed?
        high = next_power_of_2(self.size_hint)
        default = min(high, 4096)
        value = fn(BlockSizeFragment(low, high, default))
        assert isinstance(value, int)
        if value >= self.size_hint:
            return None  # max size becomes persistent reduction
        return value

    def _normalize(self, name: str, value: object) -> int | None:
        if value is None:
            return None
        return super()._normalize(name, value)

    def _fill_missing(self) -> None:
        return None


def _product(seq: Sequence[int]) -> int:
    """Return the product of the elements in the sequence."""
    return functools.reduce(operator.mul, seq, 1)



from __future__ import annotations

import collections
import contextlib
import dataclasses
import functools
from itertools import starmap
import logging
import math
from math import inf
from multiprocessing import connection
import re
import sys
import time
from typing import TYPE_CHECKING
from typing import NamedTuple

from torch._inductor.runtime.triton_compat import OutOfResources
from torch._inductor.runtime.triton_compat import PTXASError
import torch.multiprocessing as mp
from triton.testing import do_bench

from .. import exc
from ..runtime.precompile_shim import already_compiled
from .config_generation import ConfigGeneration
from .config_generation import FlatConfig
from .logger import LambdaLogger

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ..runtime.config import Config
    from ..runtime.kernel import BoundKernel
    from ..runtime.kernel import CompiledConfig
    from ..runtime.settings import Settings
    from . import ConfigSpec

_expected_errors_regexp: re.Pattern[str] = re.compile(
    # pyre-ignore[6]
    r"|".join(map(re.escape, ["[CUDA]: invalid argument"]))
)


class BaseSearch:
    """
    Base class for search algorithms. This class defines the interface and utilities for all
    search algorithms.

    Attributes:
        kernel (BoundKernel): The kernel to be tuned.
        settings (Settings): The settings associated with the kernel.
        config_spec (ConfigSpec): The configuration specification for the kernel.
        args (Sequence[object]): The arguments to be passed to the kernel.
        counters (collections.Counter): A counter to track various metrics during the search.
    """

    # pyre-fixme[11]: BoundKernel undefined?
    def __init__(self, kernel: BoundKernel, args: Sequence[object]) -> None:
        """
        Initialize the BaseSearch object.

        :param kernel: The kernel to be tuned.
        :type kernel: BoundKernel
        :param args: The arguments to be passed to the kernel.
        :type args: Sequence[object]
        """
        super().__init__()
        self.kernel = kernel
        self.settings: Settings = kernel.settings
        self.config_spec: ConfigSpec = kernel.config_spec
        self.args = args
        self.counters: collections.Counter[str] = collections.Counter()
        self.log = LambdaLogger(self.settings.autotune_log_level)

    def benchmark(self, config: Config) -> float:
        """
        Benchmark a specific configuration.

        This method compiles the kernel with the given configuration and measures its performance.

        :param config: The configuration to benchmark.
        :type config: Config
        :return: The performance of the configuration in seconds.
        :rtype: float
        """
        fn = self.kernel.compile_config(config, allow_print=False)
        if self.start_precompile_and_check_for_hangs(config, fn)():
            return self.benchmark_function(config, fn)
        return inf

    def benchmark_function(self, config: Config, fn: CompiledConfig) -> float:
        """
        Benchmark a compiled function.  This function is called by the autotuner to measure the
        performance of a specific configuration.

        :param config: The configuration to benchmark.
        :param fn: A precompiled version of config.
        :return: The performance of the configuration in seconds.
        """
        self.counters["benchmark"] += 1
        self.log.debug(lambda: f"Running benchmark for {config!r}")
        try:
            # TODO(jansel): early exit with fewer trials if early runs are slow
            t0 = time.perf_counter()
            fn(*self.args)  # make sure the kernel is compiled
            t1 = time.perf_counter()
            res = do_bench(
                functools.partial(fn, *self.args),
                return_mode="median",
            )
            t2 = time.perf_counter()
            self.log.debug(
                lambda: f"result: {res:.4f}ms (took {t1 - t0:.1f}s + {t2 - t1:.1f}s)",
            )
            return res
        except OutOfResources:
            self.log.debug("Benchmarking failed: OutOfResources")
        except PTXASError:
            self.log.warning(f"PTXASError compiling config: {config}")
        except Exception as e:
            if not _expected_errors_regexp.search(str(e)):
                raise exc.TritonError(f"{type(e).__qualname__}: {e}", config) from e
            self.log.debug(f"Benchmarking failed: {type(e).__name__}: {e}")
        return inf

    def start_precompile_and_check_for_hangs(
        self, config: Config, fn: CompiledConfig
    ) -> PrecompileFuture:
        """
        Unfortunately, Triton can hang when compiling a kernel. This function tries to
        compile the kernel with the given configuration and checks if it hangs in a subprocess.
        We also do this in parallel (when called from parallel_benchmark) to do faster autotuning.
        Note that we compile in parallel, but we benchmark one-by-one to avoid noisy results.

        :param config: The config that generated fn.
        :param fn: The function to be precompiled.
        :return: True if the compilation was successful, False if it hung.
        """
        if not self.settings.autotune_precompile:
            return PrecompileFuture.skip(self, config, True)
        ctx = mp.get_context("fork")
        # pyre-ignore[16]
        precompiler = fn.make_precompiler(*self.args)
        if precompiler is already_compiled:
            return PrecompileFuture.skip(self, config, True)
        process: mp.Process = ctx.Process(target=precompiler)
        process.start()
        return PrecompileFuture(
            search=self,
            config=config,
            process=process,
            timeout=self.settings.autotune_compile_timeout,
        )

    def parallel_benchmark(self, configs: list[Config]) -> list[tuple[Config, float]]:
        """
        Benchmark multiple configurations in parallel.

        :param configs: A list of configurations to benchmark.
        :type configs: list[Config]
        :return: A list of tuples containing configurations and their performance.
        :rtype: list[tuple[Config, float]]
        """
        fns = [self.kernel.compile_config(c, allow_print=False) for c in configs]
        if self.settings.autotune_precompile:
            is_workings = PrecompileFuture.wait_for_all(
                [
                    *starmap(
                        self.start_precompile_and_check_for_hangs,
                        zip(configs, fns, strict=True),
                    )
                ]
            )
        else:
            is_workings = [True] * len(configs)
        results = []
        for config, fn, is_working in zip(configs, fns, is_workings, strict=True):
            if is_working:
                # benchmark one-by-one to avoid noisy results
                results.append((config, self.benchmark_function(config, fn)))
            else:
                results.append((config, inf))
        return results

    def autotune(self) -> Config:
        """
        Perform autotuning to find the best configuration.

        This method searches for the optimal configuration by benchmarking multiple configurations.

        :return: The best configuration found during autotuning.
        :rtype: Config
        """
        start = time.perf_counter()
        self.log.reset()
        best = self._autotune()
        end = time.perf_counter()
        self.log(
            f"Autotuning complete in {end - start:.1f}s after searching {self.counters['benchmark']} configs.\n"
            "One can hardcode the best config and skip autotuning with:\n"
            f"    @helion.kernel(config={best!r})\n",
            level=logging.INFO + 5,
        )
        if self.settings.print_output_code:
            triton_code = self.kernel.to_triton_code(best)
            print(triton_code, file=sys.stderr)
        return best

    def _autotune(self) -> Config:
        """
        Abstract method to perform the actual autotuning.

        This method must be implemented by subclasses.

        :raises NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError


class PopulationMember(NamedTuple):
    """
    Represents a member of the population in population-based search algorithms.

    Attributes:
        perf (float): The performance of the configuration.
        flat_values (FlatConfig): The flat representation of the configuration values.
        config (Config): The full configuration object.
    """

    perf: float
    flat_values: FlatConfig
    config: Config


def performance(member: PopulationMember) -> float:
    """
    Retrieve the performance of a population member.  Used as a sort key.

    :param member: The population member.
    :type member: PopulationMember
    :return: The performance of the member.
    :rtype: float
    """
    return member.perf


class PopulationBasedSearch(BaseSearch):
    """
    Base class for search algorithms that use a population of configurations.

    Attributes:
        population (list[PopulationMember]): The current population of configurations.
        flat_spec (list[ConfigSpecFragment]): The flattened configuration specification.
    """

    def __init__(
        self,
        kernel: BoundKernel,
        args: Sequence[object],
    ) -> None:
        """
        Initialize the PopulationBasedSearch object.

        :param kernel: The kernel to be tuned.
        :type kernel: BoundKernel
        :param args: The arguments to be passed to the kernel.
        :type args: Sequence[object]
        """
        super().__init__(kernel, args)
        self.population: list[PopulationMember] = []
        self.config_gen: ConfigGeneration = ConfigGeneration(self.config_spec)

    @property
    def best(self) -> PopulationMember:
        """
        Retrieve the best configuration in the population.

        :return: The best population member.
        :rtype: PopulationMember
        """
        return min(self.population, key=performance)

    def benchmark_flat(self, flat_values: FlatConfig) -> PopulationMember:
        """
        Benchmark a flat configuration.

        :param flat_values: The flat configuration values.
        :type flat_values: FlatConfig
        :return: A population member with the benchmark results.
        :rtype: PopulationMember
        """
        config = self.config_gen.unflatten(flat_values)
        return PopulationMember(self.benchmark(config), flat_values, config)

    def parallel_benchmark_flat(
        self, to_check: list[FlatConfig]
    ) -> list[PopulationMember]:
        """
        Benchmark multiple flat configurations in parallel.

        :param to_check: A list of flat configurations to benchmark.
        :type to_check: list[FlatConfig]
        :return: A list of population members with the benchmark results.
        :rtype: list[PopulationMember]
        """
        configs = [*map(self.config_gen.unflatten, to_check)]
        result = []
        for flat_values, config_in, (config_out, perf) in zip(
            to_check, configs, self.parallel_benchmark(configs), strict=True
        ):
            assert config_in is config_out
            result.append(PopulationMember(perf, flat_values, config_in))
        return result

    def statistics(self) -> str:
        """
        Generate statistics for the current population.

        :return: A string summarizing the population performance.
        :rtype: str
        """
        return population_statistics(self.population)


def population_statistics(population: list[PopulationMember]) -> str:
    """
    Create a summary of the population performance.

    :param population: The population of configurations.
    :type population: list[PopulationMember]
    :return: A string summarizing the performance of the population.
    :rtype: str
    """
    population = sorted(population, key=performance)
    if math.isinf(population[-1].perf):
        working = [x for x in population if not math.isinf(x.perf)]
        return (
            f"failed={len(population) - len(working)} "
            f"min={working[0].perf:.4f} "
            f"mid={working[len(working) // 2].perf:.4f} "
            f"max={working[-1].perf:.4f} "
            f"best={population[0].config!s}"
        )
    return (
        f"min={population[0].perf:.4f} "
        f"mid={population[len(population) // 2].perf:.4f} "
        f"max={population[-1].perf:.4f} "
        f"best={population[0].config!s}"
    )


@dataclasses.dataclass
class PrecompileFuture:
    """
    Wraps a child process where we are precompiling a kernel.

    Attributes:
        search (BaseSearch): The search object that initiated the precompilation.
        config (Config): The configuration to be precompiled.
        process (mp.Process | None): The process running the precompilation.
        timeout (float): The timeout for the precompilation.
        start_time (float): The time when the precompilation started.
        end_time (float | None): The time when the precompilation ended.
        ok (bool | None): The result of the precompilation (True if successful, False otherwise).
    """

    search: BaseSearch
    config: Config
    process: mp.Process | None
    timeout: float
    start_time: float = dataclasses.field(default_factory=time.time)
    end_time: float | None = None
    ok: bool | None = None

    @property
    def elapsed(self) -> float:
        """Return the elapsed time since the start of the precompilation."""
        if self.end_time is not None:
            return self.end_time - self.start_time
        return time.time() - self.start_time

    def seconds_left(self) -> float:
        """Return the number of seconds left before the timeout."""
        if self.end_time is not None:
            return 0
        return self.timeout - (time.time() - self.start_time)

    def is_alive(self) -> bool:
        """Check if the precompilation process is still alive."""
        if (p := self.process) is None:
            return False
        return p.is_alive()

    @staticmethod
    def skip(search: BaseSearch, config: Config, ok: bool) -> PrecompileFuture:
        """Dummy precompile future that is already done."""
        ts = time.time()
        return PrecompileFuture(
            search=search,
            config=config,
            process=None,
            timeout=0,
            ok=ok,
            start_time=ts,
            end_time=ts,
        )

    def __call__(self) -> bool:
        """Wait for the precompilation to finish and return true on success."""
        if self.ok is not None:
            return self.ok
        process = self.process
        assert process is not None
        try:
            process.join(self.seconds_left())
        finally:
            self._mark_complete()
        assert self.ok is not None
        return self.ok

    @staticmethod
    def wait_for_all(
        futures: list[PrecompileFuture],
    ) -> list[bool]:
        """
        Wait for all precompile futures to complete.

        :param futures: A list of PrecompileFuture objects.
        :type futures: Sequence[PrecompileFuture]
        """
        remaining = [f for f in futures if f.ok is None]
        try:
            while remaining:
                remaining = PrecompileFuture._wait_for_all_step(remaining)
        except Exception:
            for f in remaining:
                if (p := f.process) is not None:
                    with contextlib.suppress(Exception):
                        p.terminate()
            raise
        result = []
        for f in futures:
            assert f.ok is not None
            result.append(f.ok)
        return result

    @staticmethod
    def _wait_for_all_step(
        futures: list[PrecompileFuture],
    ) -> list[PrecompileFuture]:
        """Wait for at least one precompile future to finish, and return the remaining ones."""
        connection.wait(
            # pyre-ignore[16]
            [f.process.sentinel for f in futures],
            min([f.seconds_left() for f in futures]),
        )
        remaining = []
        for f in futures:
            if not f.is_alive() or f.seconds_left() <= 0:
                f._mark_complete()
            else:
                remaining.append(f)
        return remaining

    def _mark_complete(self) -> bool:
        """
        Mark the precompile future as complete and kill the process if needed.

        :return: True if the precompilation was successful, False otherwise.
        """
        self.end_time = time.time()
        process = self.process
        assert process is not None
        if not process.is_alive():
            self.ok = True
            return True
        process.terminate()
        process.join(10)
        msg = f"Timeout after {self.elapsed:.0f}s compiling {self.config}"
        if process.is_alive():
            self.search.log.warning(
                msg,
                "(SIGKILL required)",
            )
            process.kill()
            process.join()
        else:
            self.search.log.warning(msg)

        self.ok = False
        return False



from __future__ import annotations

import functools
import itertools
import operator
import random
from typing import TYPE_CHECKING
from typing import cast

from helion._compat import warps_to_threads
from helion.autotuner.config_fragment import Category
from helion.autotuner.config_fragment import ConfigSpecFragment
from helion.autotuner.config_fragment import PowerOfTwoFragment

if TYPE_CHECKING:
    from helion import Config
    from helion.autotuner import ConfigSpec

FlatConfig = list[object]


class ConfigGeneration:
    def __init__(
        self,
        config_spec: ConfigSpec,
    ) -> None:
        def _collect_spec(spec: ConfigSpecFragment) -> object:
            """
            Collect a configuration specification fragment.

            :param spec: The configuration specification fragment.
            :type spec: ConfigSpecFragment
            :return: The default value of the fragment.
            :rtype: object
            """
            self.flat_spec.append(spec)
            return spec.default()

        super().__init__()
        self.config_spec = config_spec
        self.flat_spec: list[ConfigSpecFragment] = []
        config_spec.flat_config(_collect_spec)
        assert self.flat_spec, "No config values to tune"
        self.block_size_indices: list[int] = [
            i
            for i, spec in enumerate(self.flat_spec)
            if spec.category() == Category.BLOCK_SIZE
        ]
        self.num_warps_index: int = next(
            i
            for i, spec in enumerate(self.flat_spec)
            if spec.category() == Category.NUM_WARPS
        )
        self.min_block_size: int = max(
            [spec.min_size for spec in config_spec.block_sizes]
        )

    def unflatten(self, flat_values: FlatConfig) -> Config:
        """
        Convert a flat configuration back into a full configuration.

        :param flat_values: The flat configuration values.
        :type flat_values: FlatConfig
        :return: The full configuration object.
        :rtype: Config
        """

        def get_next_value(spec: ConfigSpecFragment) -> object:
            i = next(count)
            assert type(self.flat_spec[i]) is type(spec)
            return flat_values[i]

        assert len(flat_values) == len(self.flat_spec)
        count: itertools.count[int] = itertools.count()
        config = self.config_spec.flat_config(get_next_value)
        assert next(count) == len(flat_values)
        return config

    def block_numel(self, flat_config: FlatConfig) -> int:
        return functools.reduce(
            operator.mul, [cast("int", flat_config[i]) for i in self.block_size_indices]
        )

    def shrink_config(
        self, flat_config: FlatConfig, max_elements_per_thread: int
    ) -> None:
        """
        Fully random configs tend to run out of resources and tile a long time to compile.
        Here we shrink the config to a reasonable size.

        :param flat_config: config to mutate in place
        :param max_elements_per_thread: maximum number of elements per thread
        """
        num_threads = warps_to_threads(cast("int", flat_config[self.num_warps_index]))
        max_elements = max_elements_per_thread * num_threads
        while self.block_numel(flat_config) > max_elements:
            changes = 0
            for i in self.block_size_indices:
                val = flat_config[i]
                assert isinstance(val, int)
                if val > self.min_block_size:
                    flat_config[i] = val // 2
                    changes += 1
            if changes == 0:
                break

    def default_flat(self) -> FlatConfig:
        """
        Retrieve the default flat configuration.

        :return: The default flat configuration values.
        :rtype: FlatConfig
        """
        return [spec.default() for spec in self.flat_spec]

    def random_flat(self) -> FlatConfig:
        """
        Generate a random flat configuration.

        :return: A random flat configuration.
        :rtype: FlatConfig
        """
        config = [spec.random() for spec in self.flat_spec]
        self.shrink_config(config, PowerOfTwoFragment(1, 2048, 32).random())
        return config

    def random_config(self) -> Config:
        return self.unflatten(self.random_flat())

    def random_population_flat(self, n: int) -> list[FlatConfig]:
        return [self.default_flat(), *[self.random_flat() for _ in range(n - 1)]]

    def random_population(self, n: int) -> list[Config]:
        return [*map(self.unflatten, self.random_population_flat(n))]

    def differential_mutation(
        self,
        x: FlatConfig,
        a: FlatConfig,
        b: FlatConfig,
        c: FlatConfig,
        crossover_rate: float,
    ) -> FlatConfig:
        """
        The main op in differential evolution, randomly combine `x` with `a + (b - c)`.
        """
        crossover_mask = [random.random() < crossover_rate for _ in self.flat_spec]
        crossover_mask[random.randrange(len(crossover_mask))] = True
        result = [*x]
        for i, crossover in enumerate(crossover_mask):
            if crossover:
                result[i] = self.flat_spec[i].differential_mutation(a[i], b[i], c[i])
        # TODO(jansel): can this be larger? (too large and Triton compile times blow up)
        self.shrink_config(result, 8192)
        return result



from __future__ import annotations

from typing import TYPE_CHECKING

import torch
from torch._inductor.utils import triton_type

from .._compiler.ast_extension import expr_from_string
from .._compiler.compile_environment import CompileEnvironment
from ..exc import NotInsideKernel
from . import _decorators

if TYPE_CHECKING:
    import ast

    from .._compiler.inductor_lowering import CodegenState

__all__ = ["full", "zeros"]


def zeros(shape: list[object], dtype: torch.dtype = torch.float32) -> torch.Tensor:
    """
    Return a device-tensor filled with zeros

    :param shape: a list of sizes (or tile indices which are implicitly converted to sizes)
    :param dtype: torch.dtype, default is torch.float32
    :return: a device tensor of the given shape and dtype
    """
    return full(shape, 0.0 if dtype.is_floating_point else 0, dtype=dtype)


@_decorators.api(tiles_as_sizes=True)
def full(
    shape: list[object], value: float, dtype: torch.dtype = torch.float32
) -> torch.Tensor:
    """
    Create a device-tensor filled with a specified value.

    :param shape: A list of sizes (or tile indices which are implicitly converted to sizes).
    :param value: The value to fill the tensor with.
    :param dtype: The data type of the tensor, default is torch.float32.
    :return: A device tensor of the given shape and dtype.
    """
    raise NotInsideKernel


@_decorators.register_fake(full)
def _full_fake(
    shape: list[int | torch.SymInt],
    value: float,
    dtype: torch.dtype = torch.float32,
) -> torch.Tensor:
    if not isinstance(shape, (list, tuple)):
        raise TypeError(f"Expected list[SymInt], got {type(shape).__name__}")
    env = CompileEnvironment.current()
    env.add_kernel_tensor_size(shape)
    return torch.empty(
        [*shape],
        dtype=dtype,
        device=env.device,
    )


@_decorators.codegen(full)
def _full_codegen(state: CodegenState) -> ast.AST:
    fake_value = state.fake_value
    assert isinstance(fake_value, torch.Tensor)
    shape_str = state.device_function.tile_strategy.shape_str(fake_value.size())
    type_str = triton_type(fake_value.dtype)
    value_str = state.device_function.literal_expr(state.proxy_arg(1))
    return expr_from_string(f"tl.full({shape_str}, {value_str}, {type_str})")


@_decorators.get_masked_value(full)
def _(
    node: torch.fx.Node,
) -> float | bool:
    value = node.args[1]
    assert isinstance(value, (int, float, bool))
    return value



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
import logging
import os
from typing import Callable
from typing import ParamSpec

LOG_ENV_VAR = "HELION_LOGS"

"""
Usage:

HELION_LOGS=+helion.runtime.kernel python test.py

will run test.py with helion.runtime.kernel logs enabled at logging.DEBUG level
"""


@dataclass
class LogRegistry:
    """
    This registry holds mappings for logging
    """

    # alias -> list of logs
    alias_map: dict[str, list[str]] = field(default_factory=dict)

    # log name -> log level
    log_levels: dict[str, int] = field(default_factory=dict)


_LOG_REGISTRY = LogRegistry({"all": ["helion"]})


def parse_log_value(value: str) -> None:
    """
    Given a string like "foo.bar,+baz.fizz" this function parses the string and converts
    it to a mapping of {"foo.bar": logging.INFO, "baz.fizz": logging.DEBUG}
    and updates the registry with this mapping
    """
    entries = [e.strip() for e in value.split(",") if e.strip()]
    for entry in entries:
        log_level = logging.DEBUG if entry.startswith("+") else logging.INFO
        alias = entry.lstrip("+")

        if alias in _LOG_REGISTRY.alias_map:
            for log in _LOG_REGISTRY.alias_map[alias]:
                _LOG_REGISTRY.log_levels[log] = log_level
        else:
            _LOG_REGISTRY.log_levels[alias] = log_level


def init_logs_from_string(value: str) -> None:
    """
    Installs basic logging based on the input value
    """
    parse_log_value(value)

    for logger_name, level in _LOG_REGISTRY.log_levels.items():
        logger = logging.getLogger(logger_name)
        logger.setLevel(level)

        handler = logging.StreamHandler()
        handler.setLevel(level)
        handler.setFormatter(
            logging.Formatter(
                fmt=f"%(asctime)s [{logger_name}] %(levelname)s: %(message)s",
                datefmt="%Y-%m-%d %H:%M:%S",
            )
        )

        logger.addHandler(handler)
        logger.propagate = False


def init_logs() -> None:
    init_logs_from_string(os.environ.get(LOG_ENV_VAR, ""))


P = ParamSpec("P")


class LazyString:
    def __init__(
        self, func: Callable[P, str], *args: P.args, **kwargs: P.kwargs
    ) -> None:
        self.func: Callable[P, str] = func
        self.args: tuple[object, ...] = args
        self.kwargs: object = kwargs

    def __str__(self) -> str:
        return self.func(*self.args, **self.kwargs)



from __future__ import annotations

import ast
from typing import TYPE_CHECKING

import torch
from torch._inductor.codegen.simd import constant_repr
from torch.fx import has_side_effect

from .. import exc
from .._compiler.ast_extension import expr_from_string
from .._compiler.indexing_strategy import SubscriptIndexing
from . import _decorators

if TYPE_CHECKING:
    from .._compiler.inductor_lowering import CodegenState

__all__ = ["atomic_add", "load", "store"]


@has_side_effect
@_decorators.api(tiles_as_sizes=True)
def store(
    tensor: torch.Tensor,
    index: list[object],
    value: torch.Tensor | torch.SymInt | float,
    extra_mask: torch.Tensor | None = None,
) -> None:
    """Store a value from to tensor using a list of indices.

    Args:
        tensor: The tensor to load from
        index: The indices to use to index into the tensor
        value: The value to store
        extra_mask: The extra mask (beyond automatic tile bounds masking) to apply to the tensor
    Returns:
        torch.Tensor: The loaded value
    """
    raise exc.NotInsideKernel


@_decorators.prepare_args(store)
def _(
    tensor: torch.Tensor,
    index: list[object],
    value: torch.Tensor | torch.SymInt | float,
    extra_mask: torch.Tensor | None = None,
) -> tuple[torch.Tensor, list[object], torch.Tensor | torch.SymInt | int | float]:
    from helion.language.tile_proxy import Tile

    if hasattr(value, "dtype") and value.dtype != tensor.dtype:
        value = value.to(tensor.dtype)
    index = Tile._tiles_to_sizes(index)
    return (tensor, index, value, extra_mask)


@_decorators.register_fake(store)
def _(
    tensor: torch.Tensor,
    index: list[object],
    value: torch.Tensor | torch.SymInt | float,
    extra_mask: torch.Tensor | None = None,
) -> None:
    return None


@_decorators.codegen(store)
def _(state: CodegenState) -> ast.AST:
    tensor = state.proxy_arg(0)
    assert isinstance(tensor, torch.Tensor)
    subscript = state.proxy_arg(1)
    assert isinstance(subscript, (list, tuple))
    value = state.ast_arg(2)
    extra_mask = state.ast_args[3]
    assert isinstance(extra_mask, (type(None), ast.AST))
    return state.device_function.indexing_strategy.codegen_store(
        state, tensor, [*subscript], value, extra_mask
    )


@_decorators.api(tiles_as_sizes=True)
def load(
    tensor: torch.Tensor, index: list[object], extra_mask: torch.Tensor | None = None
) -> torch.Tensor:
    """Load a value from a tensor using a list of indices.

    Args:
        tensor: The tensor to load from
        index: The indices to use to index into the tensor
        extra_mask: The extra mask (beyond automatic tile bounds masking) to apply to the tensor
    Returns:
        torch.Tensor: The loaded value
    """
    raise exc.NotInsideKernel


@_decorators.register_fake(load)
def _(
    tensor: torch.Tensor, index: list[object], extra_mask: torch.Tensor | None = None
) -> torch.Tensor:
    return tensor.new_empty(SubscriptIndexing.compute_shape(tensor, index))


@_decorators.codegen(load)
def _(state: CodegenState) -> ast.AST:
    tensor = state.proxy_arg(0)
    assert isinstance(tensor, torch.Tensor)
    subscript = state.proxy_arg(1)
    assert isinstance(subscript, (list, tuple))
    extra_mask = state.ast_args[2]
    assert isinstance(extra_mask, (type(None), ast.AST))
    return state.device_function.indexing_strategy.codegen_load(
        state, tensor, [*subscript], extra_mask
    )


@_decorators.get_masked_value(load)
def _(node: torch.fx.Node) -> int:
    return 0  # loads are always masked to 0


@has_side_effect
@_decorators.api()
def atomic_add(
    target: torch.Tensor,
    index: list[object],
    value: torch.Tensor | float,
    sem: str = "relaxed",
) -> None:
    """
    Atomically add a value to a target tensor.

    Args:
        target: The tensor to add to
        index: Indices into target for way to accumulate values
        value: The value to add
        sem: The memory ordering semantics (default: 'relaxed')

    Returns:
        None
    """
    raise exc.NotInsideKernel


@_decorators.prepare_args(atomic_add)
def _(
    target: torch.Tensor,
    index: list[object],
    value: torch.Tensor | float,
    sem: str = "relaxed",
) -> tuple[torch.Tensor, object, torch.Tensor | float | int, str]:
    from helion.language.tile_proxy import Tile

    valid_sems = {"relaxed", "acquire", "release", "acq_rel"}
    if sem not in valid_sems:
        raise ValueError(
            f"Invalid memory semantic '{sem}'. Must be one of {valid_sems}."
        )

    index = Tile._prepare_index(index)
    index = Tile._tiles_to_sizes(index)

    return (target, index, value, sem)


@_decorators.register_fake(atomic_add)
def _(
    target: torch.Tensor, index: list[object], value: torch.Tensor, sem: str = "relaxed"
) -> None:
    return None


@_decorators.codegen(atomic_add)
def _(state: CodegenState) -> ast.AST:
    target = state.proxy_arg(0)
    index = state.proxy_arg(1)
    sem = expr_from_string(repr(state.proxy_arg(3)))

    assert isinstance(target, torch.Tensor)
    assert isinstance(index, list)

    indices = SubscriptIndexing.create(state, target, index)
    name = state.device_function.tensor_arg(target).name

    value_expr = state.ast_args[2]
    if isinstance(value_expr, (int, float, bool)):
        value_expr = expr_from_string(constant_repr(value_expr))
    assert isinstance(value_expr, ast.AST)
    return expr_from_string(
        f"tl.atomic_add({name} + offset, value, mask=mask, sem=sem)",
        value=value_expr,
        offset=indices.index_expr,
        mask=indices.mask_expr,
        sem=sem,
    )



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING
from typing import Callable
from typing import MutableSequence
from typing import TypeVar

from torch.fx.node import map_aggregate

from ..exc import InvalidConfig
from .config_fragment import ConfigSpecFragment
from .config_fragment import assert_integer_power_of_two

if TYPE_CHECKING:
    from . import ConfigSpec

    _T = TypeVar("_T")
    _D = TypeVar("_D")


@dataclasses.dataclass
class _BlockIdItem:
    # the block_ids used in the IR
    block_ids: list[int]

    @property
    def block_id(self) -> int:
        """Return the first block_id for this item."""
        return self.block_ids[0]

    def _fill_missing(self) -> object:
        """Provide a value when not provided by the user."""
        raise NotImplementedError

    def _normalize(self, name: str, value: object) -> object:
        """Validate and normalize the value for this item."""
        raise NotImplementedError

    def _fragment(self, base: ConfigSpec) -> ConfigSpecFragment:
        """Return the fragment used for autotunging for this item."""
        raise NotImplementedError

    def _flat_config(
        self, base: ConfigSpec, fn: Callable[[ConfigSpecFragment], object]
    ) -> object:
        return fn(self._fragment(base))


_BlockIdItemT = TypeVar("_BlockIdItemT", bound=_BlockIdItem)


class BlockIdSequence(MutableSequence[_BlockIdItemT]):
    """
    A sequence of _BlockIdItem subclasses that allows for efficient
    mapping from block_id to index in the sequence.  A generic data
    structure used to store different types of configuration specs.
    """

    def __init__(self) -> None:
        self._data: list[_BlockIdItemT] = []
        self._block_id_to_index: dict[int, int] = {}

    def __len__(self) -> int:
        return len(self._data)

    def _reindex(self) -> None:
        """Rebuild the mapping from block_id to index."""
        new_index = {}
        for i, item in enumerate(self._data):
            for block_id in item.block_ids:
                new_index[block_id] = i
        self._block_id_to_index = new_index

    def __getitem__(self, index: int) -> _BlockIdItemT:
        return self._data[index]

    def __setitem__(self, index: int, value: _BlockIdItemT) -> None:
        self._data[index] = value
        self._reindex()  # could be faster, but uncommon case

    def __delitem__(self, index: int) -> None:
        del self._data[index]
        self._reindex()  # could be faster, but uncommon case

    def clear(self) -> None:
        self._data.clear()
        self._block_id_to_index.clear()

    def append(self, value: _BlockIdItemT) -> None:
        """Append a new item to the end of the sequence."""
        index = len(self._data)
        self._data.append(value)
        for block_id in value.block_ids:
            self._block_id_to_index[block_id] = index

    def insert(self, index: int, value: _BlockIdItemT) -> None:
        """Insert a new item at the given index."""
        if index == len(self._data):
            self.append(value)
            return
        self._data.insert(index, value)
        self._reindex()  # could be faster, but uncommon case

    def block_id_to_index(self, block_id: int) -> int:
        """Return the index of the block_id in the config."""
        return self._block_id_to_index[block_id]

    def block_id_lookup(self, block_id: int) -> _BlockIdItemT:
        """Return the index of the block_id in the config."""
        return self._data[self._block_id_to_index[block_id]]

    def disable_block_id(self, block_id: int) -> None:
        """Remove configuration choice for the given block_id."""
        self._data = [x for x in self._data if block_id not in x.block_ids]
        self._reindex()

    def config_get(
        self, config: list[_T], block_id: int, default: _D = None
    ) -> _T | _D:
        """
        Get the config value for the given block_id, or return default if not found.
        """
        index = self._block_id_to_index.get(block_id, None)
        if index is None:
            return default
        return config[index]

    def _flat_config(
        self, base: ConfigSpec, fn: Callable[[ConfigSpecFragment], object]
    ) -> list[object]:
        """Map a flattened version of the config using the given function."""
        return [spec._flat_config(base, fn) for spec in self._data]

    def _normalize(
        self, name: str, values: object, *, flatten: bool = False
    ) -> list[object]:
        """Validate and normalize the values for this config item."""
        if flatten:
            if values is None:
                values = ()
            new_values = []
            # pyre-ignore[6]
            map_aggregate(values, new_values.append)
            values = new_values
        elif not isinstance(values, (list, tuple, type(None))):
            raise InvalidConfig(
                f"Unexpected type for config[{name!r}], expected list or None, got {type(values).__name__}"
            )
        values = [*(values or ())]
        size = len(self)
        if len(values) > size:
            raise InvalidConfig(
                f"Too many values for config[{name!r}], expected {size}, got {len(values)}"
            )
        if len(values) < size:
            try:
                for spec in self._data[len(values) :]:
                    values.append(spec._fill_missing())
            except NotImplementedError:
                raise InvalidConfig(
                    f"Not enough values for config[{name!r}]: expected {size} block sizes "
                    f"(one for each tiled dimension), got {len(values)}. "
                    f"Did you forget to specify block sizes for all your hl.tile() dimensions?"
                ) from None
        for i, spec in enumerate(self._data):
            values[i] = spec._normalize(f"config[{name}][{i}]", values[i])
        return values

    def _remove_duplicates(self) -> None:
        new_specs = []
        for spec in self:
            other = self.block_id_lookup(spec.block_id)
            if other is spec:
                new_specs.append(spec)
            elif len(spec.block_ids) != len(other.block_ids):
                # this will cause invalid config errors with loop orders
                # remove them both
                self.disable_block_id(spec.block_id)
                self._remove_duplicates()  # start over
                return
        if len(new_specs) != len(self):
            self._data = new_specs
            self._reindex()


class _PowerOfTwoBlockIdItem(_BlockIdItem):
    def _normalize(self, name: str, value: object) -> int | None:
        try:
            return assert_integer_power_of_two(value)
        except InvalidConfig:
            raise InvalidConfig(
                f"{name} must be a power of two, got {value!r}"
            ) from None



from __future__ import annotations

import re
import sys
import threading
import traceback
import typing
from typing import TYPE_CHECKING
from typing import Protocol
from typing import TypeVar

from torch.fx.traceback import get_current_meta
from torch.fx.traceback import has_preserved_node_meta

from .traceback_compat import format_frame_summary

if TYPE_CHECKING:
    import ast
    from typing_extensions import Self

    from .ast_extension import ExtendedAST

    class _TLS(Protocol):
        locations: list[SourceLocation]

    _T = TypeVar("_T", ast.AST, ExtendedAST)


# pyre-ignore-all-errors[16, 28]: lineno/colno/etc are not defined
tls: _TLS = typing.cast("_TLS", threading.local())


class SourceLocation(traceback.FrameSummary):
    """Represents a location in the source code a node came from."""

    def __init__(
        self,
        lineno: int,
        colno: int,
        end_lineno: int,
        end_colno: int,
        name: str,
        filename: str,
    ) -> None:
        if sys.version_info >= (3, 11):
            super().__init__(
                lookup_line=False,
                lineno=lineno,
                end_lineno=end_lineno,
                colno=colno,
                end_colno=end_colno,
                name=name,
                filename=filename,
            )
        else:
            super().__init__(
                lookup_line=False,
                lineno=lineno,
                name=name,
                filename=filename,
            )
            self.end_lineno = end_lineno
            self.colno = colno
            self.end_colno = end_colno

    @staticmethod
    def from_ast(node: ast.AST) -> SourceLocation:
        from .host_function import HostFunction

        host_function = HostFunction.current()
        code = host_function.fn.__code__
        offset = code.co_firstlineno - 1
        return SourceLocation(
            node.lineno + offset,
            node.col_offset + host_function.column_offset,
            node.end_lineno + offset,
            node.end_col_offset + host_function.column_offset,
            filename=code.co_filename,
            name=code.co_name,
        )

    def to_ast(self, node: _T) -> _T:
        if "lineno" in node._attributes:
            node.lineno = self.lineno
            node.col_offset = self.colno
            node.end_lineno = self.end_lineno
            node.end_col_offset = self.end_colno
        return node

    def __str__(self) -> str:
        return f"{self.filename}:{self.lineno}"

    def __repr__(self) -> str:
        return f"<SourceLocation {re.sub(r'^.*/', '', self.filename)}:{self.lineno}>"

    def format(self) -> str:
        return format_frame_summary(self)

    def _key(self) -> tuple[str, int | None, int, int, int]:
        return (self.filename, self.lineno, self.colno, self.end_lineno, self.end_colno)

    def __hash__(self) -> int:
        return hash(self._key())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SourceLocation):
            return False
        return self._key() == other._key()

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __lt__(self, other: Self) -> bool:
        return self._key() < other._key()

    def __le__(self, other: Self) -> bool:
        return self._key() <= other._key()

    def __gt__(self, other: Self) -> bool:
        return self._key() > other._key()

    def __ge__(self, other: Self) -> bool:
        return self._key() >= other._key()

    def __bool__(self) -> bool:
        return True

    def __enter__(self) -> None:
        try:
            tls.locations.append(self)
        except AttributeError:
            tls.locations = [self]
        self.set_fx_location()

    def __exit__(self, *args: object) -> None:
        locations = tls.locations
        locations.pop()
        if locations:
            locations[-1].set_fx_location()
        else:
            UnknownLocation().set_fx_location()

    def set_fx_location(self) -> None:
        if has_preserved_node_meta():
            meta = get_current_meta()
            meta["location"] = self
            meta["stack_trace"] = re.sub(
                r"^((?:.*\n){2})(.|\n)*$", r"\1", self.format()
            )


class UnknownLocation(SourceLocation):
    def __init__(self) -> None:
        super().__init__(0, 0, 0, 0, "<unknown>", "<unknown>")

    def __str__(self) -> str:
        return "unknown location"

    def __repr__(self) -> str:
        return "<unknown location>"

    def format(self) -> str:
        return "unknown location\n"

    def __bool__(self) -> bool:
        return False

    def set_fx_location(self) -> None:
        if has_preserved_node_meta():
            meta = get_current_meta()
            meta.pop("location", None)
            meta.pop("stack_trace", None)


def current_location() -> SourceLocation:
    try:
        return tls.locations[-1]
    except (AttributeError, IndexError):
        return UnknownLocation()



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING
from typing import NamedTuple

from helion._compiler.ast_extension import expr_from_string
from helion._compiler.ast_extension import statement_from_string
from helion._compiler.host_function import HostFunction

if TYPE_CHECKING:
    import ast

    import sympy

    from helion._compiler.inductor_lowering import CodegenState


class ProgramID(NamedTuple):
    pid_var: str
    block_size_var: str
    numel: sympy.Expr

    def host_cdiv(self) -> str:
        numel_str = HostFunction.current().sympy_expr(self.numel)
        if self.block_size_var == "1":
            return numel_str
        return f"triton.cdiv({numel_str}, {self.block_size_var})"

    def device_cdiv(self, state: CodegenState) -> str:
        numel_str = state.sympy_expr(self.numel)
        if self.block_size_var == "1":
            return numel_str
        return f"tl.cdiv({numel_str}, {self.block_size_var})"


@dataclasses.dataclass
class SharedProgramID:
    """
    Use the same PID for all blocks
    TODO(oulgen): Currently only supports 1 dimension
    """

    shared_pid_var: str
    pids: list[ProgramIDs] = dataclasses.field(default_factory=list)

    def codegen_pid_init(
        self,
    ) -> ast.stmt:
        return statement_from_string(f"{self.shared_pid_var} = tl.program_id(0)")

    def codegen_test(self, state: CodegenState) -> ast.AST:
        blocks = []
        for pid in self.pids:
            blocks.append(pid.combined_device_cdiv(state))

        assert len(blocks) > 0
        return expr_from_string(f"{self.shared_pid_var} < ({'+ '.join(blocks)})")

    def codegen(self, state: CodegenState) -> None:
        # TODO(oulgen): We need CSE between codegen_test and codegen for shared device cdivs
        blocks = []
        for pid in self.pids[:-1]:
            blocks.append(pid.combined_device_cdiv(state))

        if blocks:
            state.codegen.statements_stack[-1].insert(
                0,
                statement_from_string(
                    f"{self.shared_pid_var} -= ({'+ '.join(blocks)})"
                ),
            )

    def codegen_grid(self) -> ast.AST:
        return expr_from_string(
            f"({'+ '.join(pid.combined_host_cdiv() for pid in self.pids)},)"
        )


@dataclasses.dataclass
class ProgramIDs:
    pids: list[ProgramID] = dataclasses.field(default_factory=list)
    shared_pid_var: str | None = None

    def append(self, pid: ProgramID) -> None:
        self.pids.append(pid)

    def codegen(self, state: CodegenState) -> None:
        raise NotImplementedError

    def codegen_grid(self) -> ast.AST:
        raise NotImplementedError

    def combined_device_cdiv(self, state: CodegenState) -> str:
        raise NotImplementedError

    def combined_host_cdiv(self) -> str:
        raise NotImplementedError


class GridProgramIDs(ProgramIDs):
    """Use the cuda x/y/z launch grid for PIDs"""

    def codegen(self, state: CodegenState) -> None:
        for i, pid in enumerate(self.pids):
            state.codegen.statements_stack[-1].insert(
                i, statement_from_string(f"{pid.pid_var} = tl.program_id({i})")
            )

    def codegen_grid(self) -> ast.AST:
        assert len(self.pids) <= 3
        return expr_from_string(f"({', '.join(pid.host_cdiv() for pid in self.pids)},)")


class VirtualProgramIDs(ProgramIDs):
    """Only use the x grid and compute other dimensions"""

    def combined_device_cdiv(self, state: CodegenState) -> str:
        return " * ".join(pid.device_cdiv(state) for pid in self.pids)

    def combined_host_cdiv(self) -> str:
        return " * ".join(pid.host_cdiv() for pid in self.pids)

    def codegen(self, state: CodegenState) -> None:
        pid_var = self.shared_pid_var or "tl.program_id(0)"

        num_blocks = [
            state.device_function.new_var(f"num_blocks_{i}")
            for i in range(len(self.pids[:-1]))
        ]
        statements = [
            statement_from_string(f"{num_block} = {pid.device_cdiv(state)}")
            for num_block, pid in zip(num_blocks, self.pids[:-1], strict=True)
        ]
        for i, pid in enumerate(self.pids):
            expr = pid_var
            if i > 0:
                divisor = " * ".join(num_blocks[:i])
                expr = f"({expr}) // ({divisor})"
            if i + 1 < len(self.pids):
                expr = f"({expr}) % ({num_blocks[i]})"
            statements.append(statement_from_string(f"{pid.pid_var} = {expr}"))
        state.codegen.statements_stack[-1][:] = [
            *statements,
            *state.codegen.statements_stack[-1],
        ]

    def codegen_grid(self) -> ast.AST:
        return expr_from_string(f"({self.combined_host_cdiv()},)")


@dataclasses.dataclass
class L2GroupingProgramIDs(ProgramIDs):
    """Used grouped iteration order to promote L2 cache reuse in matmuls"""

    group_size: int = 1

    def codegen(self, state: CodegenState) -> None:
        assert len(self.pids) == 2
        new_var = state.device_function.new_var
        pid = "tl.program_id(0)"
        num_pid_m = new_var("num_pid_m")
        num_pid_n = new_var("num_pid_n")
        num_pid_in_group = new_var("num_pid_in_group")
        group_id = new_var("group_id")
        first_pid_m = new_var("first_pid_m")
        group_size_m = new_var("group_size_m")
        state.codegen.statements_stack[-1][:] = [
            statement_from_string(f"{num_pid_m} = {self.pids[0].device_cdiv(state)}"),
            statement_from_string(f"{num_pid_n} = {self.pids[1].device_cdiv(state)}"),
            statement_from_string(
                f"{num_pid_in_group} = {self.group_size} * {num_pid_n}"
            ),
            statement_from_string(f"{group_id} = {pid} // {num_pid_in_group}"),
            statement_from_string(f"{first_pid_m} = {group_id} * {self.group_size}"),
            statement_from_string(
                f"{group_size_m} = min({num_pid_m} - {first_pid_m}, {self.group_size})"
            ),
            statement_from_string(
                f"{self.pids[0].pid_var} = {first_pid_m} + (({pid} % {num_pid_in_group}) % {group_size_m})"
            ),
            statement_from_string(
                f"{self.pids[1].pid_var} = ({pid} % {num_pid_in_group}) // {group_size_m}"
            ),
            *state.codegen.statements_stack[-1],
        ]

    def codegen_grid(self) -> ast.AST:
        return expr_from_string(
            f"({' * '.join(pid.host_cdiv() for pid in self.pids)},)"
        )



from __future__ import annotations

from .config_fragment import BooleanFragment as BooleanFragment
from .config_fragment import EnumFragment as EnumFragment
from .config_fragment import IntegerFragment as IntegerFragment
from .config_fragment import PowerOfTwoFragment as PowerOfTwoFragment
from .config_spec import ConfigSpec as ConfigSpec
from .differential_evolution import (
    DifferentialEvolutionSearch as DifferentialEvolutionSearch,
)
from .finite_search import FiniteSearch as FiniteSearch
from .random_search import RandomSearch as RandomSearch
